# Tries to evaluate the likelihood at LCDM's best fit of Planck 2015, with CAMB and CLASS
from copy import deepcopy

from .common_cosmo import body_of_test
from cobaya.cosmo_input import planck_precision

# Generating plots in Travis
import matplotlib

matplotlib.use('agg')

# Downgrade of Planck 2018 precision/model

cmb_precision = deepcopy(planck_precision)
cmb_precision["camb"].update({
    "halofit_version": "takahashi",
    "bbn_predictor": "BBN_fitting_parthenope"
})
cmb_precision["classy"].update({
    # Something changed in CLASS that makes halofit work differently that it did
    # before, and hamcode gives results closer to the original chi2, computed with halofit
    # "non linear": "halofit",
    "non linear": "hmcode",
})

# Derived parameters not understood by CLASS
# https://wiki.cosmos.esa.int/planckpla2015/images/b/b9/Parameter_tag_definitions_2015.pdf
classy_unknown = ["zstar", "rstar", "thetastar", "DAstar", "zdrag",
                  "kd", "thetad", "zeq", "keq", "thetaeq", "thetarseq",
                  "DH", "Y_p"]

# Small chi2 difference with CLASS (total still <0.5)
classy_extra_tolerance = 0.49


def test_planck_2015_t_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowl_highTT)
    best_fit.pop("H0")
    info_likelihood = lik_info_lowl_highTT
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lowl_highTT
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowl_highTT, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_p_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowTEB_highTTTEEE)
    best_fit.pop("H0")
    info_likelihood = lik_info_lowTEB_highTTTEEE
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lowTEB_highTTTEEE
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowTEB_highTTTEEE, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_l_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lensing)
    best_fit.pop("H0")
    info_likelihood = lik_info_lensing
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lensing
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lensing, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_l2_camb(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lensing)
    best_fit.pop("H0")
    lik_name = "planck_2015_lensing_cmblikes"
    clik_name = "planck_2015_lensing"
    info_likelihood = {lik_name: lik_info_lensing[clik_name]}
    chi2_lensing_cmblikes = deepcopy(chi2_lensing)
    chi2_lensing_cmblikes[lik_name] = chi2_lensing[clik_name]
    info_theory = {"camb": {"extra_args": planck_precision["camb"]}}
    best_fit_derived = derived_lensing
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lensing_cmblikes, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_t_classy(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowl_highTT)
    best_fit.pop("theta_MC_100")
    info_likelihood = lik_info_lowl_highTT
    info_theory = {"classy": {"extra_args": planck_precision["classy"]}}
    best_fit_derived = deepcopy(derived_lowl_highTT)
    for p in classy_unknown:
        best_fit_derived.pop(p, None)
    chi2_lowl_highTT_classy = deepcopy(chi2_lowl_highTT)
    chi2_lowl_highTT_classy["tolerance"] += classy_extra_tolerance
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowl_highTT_classy, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_p_classy(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lowTEB_highTTTEEE)
    best_fit.pop("theta_MC_100")
    info_likelihood = lik_info_lowTEB_highTTTEEE
    info_theory = {"classy": {"extra_args": planck_precision["classy"]}}
    best_fit_derived = deepcopy(derived_lowTEB_highTTTEEE)
    for p in classy_unknown:
        best_fit_derived.pop(p, None)
    chi2_lowTEB_highTTTEEE_classy = deepcopy(chi2_lowTEB_highTTTEEE)
    chi2_lowTEB_highTTTEEE_classy["tolerance"] += classy_extra_tolerance
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lowTEB_highTTTEEE_classy, best_fit_derived,
                 skip_not_installed=skip_not_installed)


def test_planck_2015_l_classy(packages_path, skip_not_installed):
    best_fit = deepcopy(params_lensing)
    best_fit.pop("theta_MC_100")
    info_likelihood = lik_info_lensing
    info_theory = {"classy": {"extra_args": planck_precision["classy"]}}
    best_fit_derived = deepcopy(derived_lensing)
    for p in classy_unknown:
        best_fit_derived.pop(p, None)
    chi2_lensing_classy = deepcopy(chi2_lensing)
    chi2_lensing_classy["tolerance"] += classy_extra_tolerance
    body_of_test(packages_path, best_fit, info_likelihood, info_theory,
                 chi2_lensing_classy, best_fit_derived,
                 skip_not_installed=skip_not_installed)


# Temperature only #######################################################################

lik_info_lowl_highTT = {"planck_2015_lowl": None, "planck_2015_plikHM_TT": None}

chi2_lowl_highTT = {"planck_2015_lowl": 15.39,
                    "planck_2015_plikHM_TT": 761.09,
                    "tolerance": 0.1}

params_lowl_highTT = {
    # Sampled
    "omegabh2": 0.02249139,
    "omegach2": 0.1174684,
    # only one of the next two is finally used!
    "H0": 68.43994,  # will be ignored in the CAMB case
    "theta_MC_100": 1.041189,  # will be ignored in the CLASS case
    "tau": 0.1249913,
    "logA": 3.179,
    "ns": 0.9741693,
    # Planck likelihood
    "A_planck": 1.00027,
    "A_cib_217": 61.1,
    "xi_sz_cib": 0.56,
    "A_sz": 6.84,
    "ps_A_100_100": 242.9,
    "ps_A_143_143": 43.0,
    "ps_A_143_217": 46.1,
    "ps_A_217_217": 104.1,
    "ksz_norm": 0.00,
    "gal545_A_100": 7.31,
    "gal545_A_143": 9.07,
    "gal545_A_143_217": 17.99,
    "gal545_A_217": 82.9,
    "calib_100T": 0.99796,
    "calib_217T": 0.99555}

derived_lowl_highTT = {
    # param: [best_fit, sigma]
    "H0": [params_lowl_highTT["H0"], 1.2],
    "omegal": [0.6998, 0.016],
    "omegam": [0.3002, 0.016],
    "sigma8": [0.8610, 0.023],
    "zrei": [13.76, 2.5],
    # "YHe": [0.2454462, 0.0001219630],
    # "Y_p": [0.2467729, 0.0001224069],
    # "DH": [2.568606e-5, 0.05098625e-5],
    "age": [13.7664, 0.048],
    "zstar": [1089.55, 0.52],
    "rstar": [145.00, 0.55],
    "thetastar": [1.041358, 0.0005117986],
    "DAstar": [13.924, 0.050],
    "zdrag": [1060.05, 0.52],
    "rdrag": [147.63, 0.53],
    "kd": [0.14039, 0.00053],
    "thetad": [0.160715, 0.00029],
    "zeq": [3345, 58],
    "keq": [0.010208, 0.00018],
    "thetaeq": [0.8243, 0.011],
    "thetarseq": [0.4550, 0.0058]}

# Best fit Polarization ##################################################################

lik_info_lowTEB_highTTTEEE = {"planck_2015_lowTEB": None,
                              "planck_2015_plikHM_TTTEEE": None}

chi2_lowTEB_highTTTEEE = {"planck_2015_lowTEB": 10496.93,
                          "planck_2015_plikHM_TTTEEE": 2431.65,
                          "tolerance": 0.15}

params_lowTEB_highTTTEEE = {
    # Sampled
    "omegabh2": 0.02225203,
    "omegach2": 0.1198657,
    # only one of the next two is finally used!
    "H0": 67.25,  # will be ignored in the CAMB case
    "theta_MC_100": 1.040778,  # will be ignored in the CLASS case
    "logA": 3.0929,
    "ns": 0.9647522,
    "tau": 0.07888604,
    # Planck likelihood
    "A_planck": 1.00029,
    "A_cib_217": 66.4,
    "xi_sz_cib": 0.13,
    "A_sz": 7.17,
    "ps_A_100_100": 255.0,
    "ps_A_143_143": 40.1,
    "ps_A_143_217": 36.4,
    "ps_A_217_217": 98.7,
    "ksz_norm": 0.00,
    "gal545_A_100": 7.34,
    "gal545_A_143": 8.97,
    "gal545_A_143_217": 17.56,
    "gal545_A_217": 81.9,
    "galf_EE_A_100": 0.0813,
    "galf_EE_A_100_143": 0.0488,
    "galf_EE_A_100_217": 0.0995,
    "galf_EE_A_143": 0.1002,
    "galf_EE_A_143_217": 0.2236,
    "galf_EE_A_217": 0.645,
    "galf_TE_A_100": 0.1417,
    "galf_TE_A_100_143": 0.1321,
    "galf_TE_A_100_217": 0.307,
    "galf_TE_A_143": 0.155,
    "galf_TE_A_143_217": 0.338,
    "galf_TE_A_217": 1.667,
    "calib_100T": 0.99818,
    "calib_217T": 0.99598}

derived_lowTEB_highTTTEEE = {
    # param: [best_fit, sigma]
    "H0": [params_lowTEB_highTTTEEE["H0"], 0.66],
    "omegal": [0.6844, 0.0091],
    "omegam": [0.3156, 0.0091],
    "sigma8": [0.8310, 0.013],
    "zrei": [10.07, 1.6],
    # "YHe": [0.2453409, 0.000072],
    # "Y_p": [0.2466672, 0.000072],
    # "DH": [2.6136e-5, 0.030e-5],
    "age": [13.8133, 0.026],
    "zstar": [1090.057, 0.30],
    "rstar": [144.556, 0.32],
    "thetastar": [1.040967, 0.00032],
    "DAstar": [13.8867, 0.030],
    "zdrag": [1059.666, 0.31],
    "rdrag": [147.257, 0.31],
    "kd": [0.140600, 0.00032],
    "thetad": [0.160904, 0.00018],
    "zeq": [3396.2, 33],
    "keq": [0.010365, 0.00010],
    "thetaeq": [0.8139, 0.0063],
    "thetarseq": [0.44980, 0.0032]}

# Best fit lensing (from best combination with lowTEB+_highTTTEEE ########################

lik_info_lensing = {"planck_2015_lensing": None}

chi2_lensing = {"planck_2015_lensing": 9.78, "tolerance": 0.09}

params_lensing = {
    # Sampled
    "omegabh2": 0.022274,
    "omegach2": 0.11913,
    # only one of the next two is finally used!
    "H0": 67.56,  # will be ignored in the CAMB case
    "theta_MC_100": 1.040867,  # will be ignored in the CLASS case
    "logA": 3.0600,
    "ns": 0.96597,
    "tau": 0.0639,
    # Planck likelihood
    "A_planck": 0.99995}

derived_lensing = {
    # param: [best_fit, sigma]
    "H0": [params_lensing["H0"], 0.64],
    "omegal": [0.6888, 0.0087],
    "omegam": [0.3112, 0.0087],
    "sigma8": [0.8153, 0.0087],
    "zrei": [8.64, 1.3],
    # "YHe": [0.245350, 0.000071],
    # "Y_p": [0.246677, 0.000072],
    # "DH": [2.6095e-5, 0.030e-5],
    "age": [13.8051, 0.026],
    "zstar": [1089.966, 0.29],
    "rstar": [144.730, 0.31],
    "thetastar": [1.041062, 0.00031],
    "DAstar": [13.9022, 0.029],
    "zdrag": [1059.6468, 0.31],
    "rdrag": [147.428, 0.30],
    "kd": [0.140437, 0.00032],
    "thetad": [0.160911, 0.00018],
    "zeq": [3379.1, 32],
    "keq": [0.010313, 0.000096],
    "thetaeq": [0.8171, 0.0060],
    "thetarseq": [0.45146, 0.0031]}



from cobaya.likelihoods.planck_NPIPE_highl_CamSpec.planck_NPIPE_CamSpec_python import Planck2020CamSpecPython


class TE(Planck2020CamSpecPython):
    r"""
    High-$\ell$ TE \textsc{CamSpec} PR4-12.6cl likelihood using \texttt{NPIPE} (Planck PR4) data
    \cite{rosenberg22}.
    """
    pass



# Global
import os
import pickle
from itertools import chain
import numpy as np
import re
from typing import Optional, List, Dict

# Local
from cobaya.conventions import Extension, packages_path_input
from cobaya.tools import str_to_list, get_translated_params, get_cache_path
from cobaya.parameterization import is_sampled_param
from cobaya.input import update_info
from cobaya.log import LoggedError, get_logger, is_debug

_covmats_file = "covmats_database.pkl"

log = get_logger(__name__)

covmat_folders = [
    "{%s}/data/planck_supp_data_and_covmats/covmats/" % packages_path_input,
    "{%s}/data/bicep_keck_2018/BK18_cosmomc/planck_covmats/" % packages_path_input]

# Global instance of loaded database, for fast calls to get_best_covmat in GUI
_loaded_covmats_database: Optional[List[Dict]] = None


def get_covmat_database(packages_path, cached=True) -> List[dict]:
    # Get folders with corresponding components installed
    installed_folders = [folder for folder in covmat_folders if
                         os.path.exists(
                             folder.format(**{packages_path_input: packages_path}))]
    covmats_database_fullpath = os.path.join(get_cache_path(), _covmats_file)
    # Check if there is a usable cached one
    if cached:
        try:
            with open(covmats_database_fullpath, "rb") as f:
                covmat_database = pickle.load(f)
            # quick and dirty hash for regeneration: check number of .covmat files
            num_files = len(list(chain(
                *[[filename for filename in os.listdir(
                    folder.format(**{packages_path_input: packages_path}))
                   if filename.endswith(Extension.covmat)]
                  for folder in installed_folders])))
            assert num_files == len(covmat_database)
            log.debug("Loaded cached covmats database")
            return covmat_database
        except:
            log.info("No cached covmat database present, not usable or not up-to-date. "
                     "Will be re-created and cached.")
            pass
    # Create it (again)
    covmat_database = []
    for folder in installed_folders:
        folder_full = folder.format(
            **{packages_path_input: packages_path}).replace("/", os.sep)
        for filename in os.listdir(folder_full):
            try:
                with open(os.path.join(folder_full, filename),
                          encoding="utf-8") as covmat:
                    header = covmat.readline()
                assert header.strip().startswith("#")
                params = header.strip().lstrip("#").split()
            except:
                continue
            covmat_database.append({"folder": folder, "name": filename, "params": params})
    if cached:
        with open(covmats_database_fullpath, "wb") as f:
            pickle.dump(covmat_database, f)
    return covmat_database


def get_best_covmat(info, packages_path=None, cached=True, random_state=None):
    """
    Chooses optimal covmat from a database, based on common parameters and likelihoods.

    Returns a dict `{folder: [folder_of_covmat], name: [file_name_of_covmat],
    params: [parameters_in_covmat], covmat: [covariance_matrix]}`.
    """
    packages_path = packages_path or info.get(packages_path_input)
    if not packages_path:
        raise LoggedError(log, "Needs a path to the external packages' installation.")
    updated_info = update_info(info, strict=False)
    for p, pinfo in list(updated_info["params"].items()):
        if not is_sampled_param(pinfo):
            updated_info["params"].pop(p)
    info_sampled_params = updated_info["params"]
    covmat_data = get_best_covmat_ext(packages_path, updated_info["params"],
                                      updated_info["likelihood"], random_state, cached)
    if covmat_data is None:
        return None
    covmat = np.atleast_2d(np.loadtxt(os.path.join(
        covmat_data["folder"].format(packages_path=packages_path), covmat_data["name"])))
    params_in_covmat = get_translated_params(info_sampled_params, covmat_data["params"])
    indices = [covmat_data["params"].index(p) for p in params_in_covmat.values()]
    covmat_data["covmat"] = covmat[indices][:, indices]
    covmat_data["params"] = params_in_covmat
    return covmat_data


def get_best_covmat_ext(packages_path, params_info, likelihoods_info, random_state,
                        cached=True) -> Optional[dict]:
    """
    Actual covmat finder used by `get_best_covmat`. Call directly for more control on
    the parameters used.

    Returns the same dict as `get_best_covmat`, except for the covariance matrix itself.
    """
    global _loaded_covmats_database
    covmats_database = (
            _loaded_covmats_database or get_covmat_database(packages_path, cached=cached))
    _loaded_covmats_database = covmats_database
    # Prepare params and likes aliases
    params_renames = set(chain(*[
        [p] + str_to_list(info.get("renames", [])) for p, info in
        params_info.items()]))
    likes_renames = set(chain(*[[like] + str_to_list((info or {}).get("aliases", []))
                                for like, info in likelihoods_info.items()]))
    delimiters = r"[_\.]"
    likes_regexps = [re.compile(delimiters + re.escape(_like) + delimiters)
                     for _like in likes_renames]
    # Match number of params
    score_params = (
        lambda covmat: len(set(covmat["params"]).intersection(params_renames)))
    best_p = get_best_score(covmats_database, score_params)
    if not best_p:
        log.warning(
            "No covariance matrix found including at least one of the given parameters")
        return None
    # Match likelihood names / keywords
    # No debug print here: way too many!
    score_likes = (
        lambda covmat: len([0 for r in likes_regexps if r.search(covmat["name"])]))
    best_p_l = get_best_score(best_p, score_likes)
    if is_debug(log):
        log.debug("Subset based on params + likes:\n - " +
                  "\n - ".join([b["name"] for b in best_p_l]))

    # Finally, in case there is more than one, select shortest #params and name (simpler!)
    # #params first, to avoid extended models with shorter covmat name
    def score_simpler_params(covmat):
        return -len(covmat["params"])

    best_p_l_sp = get_best_score(best_p_l, score_simpler_params)
    if is_debug(log):
        log.debug("Subset based on params + likes + fewest params:\n - " +
                  "\n - ".join([b["name"] for b in best_p_l_sp]))
    score_simpler_name = (
        lambda covmat: -len(covmat["name"].replace("_", " ").replace("-", " ").split()))
    best_p_l_sp_sn = get_best_score(best_p_l_sp, score_simpler_name)
    if is_debug(log):
        log.debug("Subset based on params + likes + fewest params + shortest name:\n - " +
                  "\n - ".join([b["name"] for b in best_p_l_sp_sn]))
    # if there is more than one (unlikely), just pick one at random
    if len(best_p_l_sp_sn) > 1:
        log.warning("WARNING: >1 possible best covmats: %r",
                    [b["name"] for b in best_p_l_sp_sn])
    random_state = np.random.default_rng(random_state)
    return best_p_l_sp_sn[random_state.choice(range(len(best_p_l_sp_sn)))].copy()


def get_best_score(covmats, score_func) -> List[dict]:
    scores = np.array([score_func(covmat) for covmat in covmats])
    i_max = np.argwhere(scores == np.max(scores)).T[0]
    return [covmats[i] for i in i_max]



"""
.. module:: bib

:Synopsis: Tools and script to get the bibliography to be cited for each component
:Author: Jesus Torrado

Inspired by a similar characteristic of
`CosmoSIS <https://bitbucket.org/joezuntz/cosmosis/wiki/Home>`_.

"""

# Global
import os
from inspect import cleandoc
from typing import List, Dict, Union

# Local
from cobaya.log import logger_setup, get_logger
from cobaya.conventions import Extension, dump_sort_cosmetic
from cobaya.tools import create_banner, warn_deprecation, similar_internal_class_names
from cobaya.component import get_component_class, ComponentNotFoundError
from cobaya.input import load_input, get_used_components
from cobaya.typing import InfoDict

# Banner defaults
_default_symbol = "="
_default_length = 80

# Cobaya's own bib info
cobaya_desc = cleandoc(r"""
The posterior has been explored/maximized/reweighted using Cobaya \cite{torrado:2020dgo}.
""")

cobaya_bib = r"""
@article{Torrado:2020dgo,
    author = "Torrado, Jesus and Lewis, Antony",
    title = "{Cobaya: Code for Bayesian Analysis of hierarchical physical models}",
    eprint = "2005.05290",
    archivePrefix = "arXiv",
    primaryClass = "astro-ph.IM",
    reportNumber = "TTK-20-15",
    doi = "10.1088/1475-7516/2021/05/057",
    journal = "JCAP",
    volume = "05",
    pages = "057",
    year = "2021"
}
""".lstrip("\n")


def get_desc_component(component, kind, info=None):
    """Extract a short description of a component, if defined."""
    cls = get_component_class(component, kind)
    return cleandoc(cls.get_desc(info) or "") + "\n"


def get_bib_component(component, kind):
    """Extract the bibliographic sources of a component, if defined."""
    cls = get_component_class(component, kind)
    lines = ((cls.get_bibtex() or "").lstrip("\n").rstrip("\n") or
             "# [no bibliography information found]")
    return lines + "\n"


def get_bib_info(*infos, logger=None):
    """
    Gathers and returns the descriptions and bibliographic sources for the components
    mentioned in ``infos``.

    ``infos`` can be input dictionaries or single component names.
    """
    if not logger:
        logger_setup()
        logger = get_logger("bib")
    used_components, component_infos = get_used_components(*infos, return_infos=True)
    descs: InfoDict = {}
    bibs: InfoDict = {}
    used_components = get_used_components(*infos)
    for kind, components in used_components.items():
        if kind is None:
            continue  # we will deal with bare component names later, to avoid repetition
        descs[kind], bibs[kind] = {}, {}
        for component in components:
            try:
                descs[kind][component] = get_desc_component(
                    component, kind, component_infos[component])
                bibs[kind][component] = get_bib_component(component, kind)
            except ComponentNotFoundError:
                sugg = similar_internal_class_names(component)
                logger.error(
                    f"Could not identify component '{component}'. "
                    f"Did you mean any of the following? {sugg} (mind capitalization!)")
                continue
    # Deal with bare component names
    for component in used_components.get(None, []):
        try:
            cls = get_component_class(component)
        except ComponentNotFoundError:
            sugg = similar_internal_class_names(component)
            logger.error(
                f"Could not identify component '{component}'. "
                f"Did you mean any of the following? {sugg} (mind capitalization!)")
            continue
        kind = cls.get_kind()
        if kind not in descs:
            descs[kind], bibs[kind] = {}, {}
        if kind in descs and component in descs[kind]:
            continue  # avoid repetition
        descs[kind][component] = get_desc_component(cls, kind)
        bibs[kind][component] = get_bib_component(cls, kind)
    descs["cobaya"] = {"cobaya": cobaya_desc}
    bibs["cobaya"] = {"cobaya": cobaya_bib}
    return descs, bibs


def pretty_repr_bib(descs, bibs):
    """
    Generates a pretty-print multi-line string from component descriptions and
    bibliographical sources.
    """
    # Sort them "optimally"
    sorted_kinds = [k for k in dump_sort_cosmetic if k in descs]
    sorted_kinds += [k for k in descs if k not in dump_sort_cosmetic]
    txt = ""
    txt += create_banner(
        "Descriptions", symbol=_default_symbol, length=_default_length) + "\n"
    for kind in sorted_kinds:
        txt += kind + ":\n\n"
        for component, desc in descs[kind].items():
            txt += " * [%s] %s\n" % (component, desc)
        txt += "\n"
    txt += "\n" + create_banner(
        "Bibtex", symbol=_default_symbol, length=_default_length) + "\n"
    for kind in sorted_kinds:
        for component, bib in bibs[kind].items():
            txt += "\n### %s " % component + "########################" + "\n\n"
            txt += bib
    return txt.lstrip().rstrip() + "\n"


# Command-line script ####################################################################

def bib_script(args=None):
    """Command line script for the bibliography."""
    warn_deprecation()
    # Parse arguments and launch
    import argparse
    parser = argparse.ArgumentParser(prog="cobaya bib", description=(
        "Prints bibliography to be cited for one or more components or input files."))
    parser.add_argument("files_or_components", action="store", nargs="+",
                        metavar="input_file.yaml|component_name",
                        help="Component(s) or input file(s) whose bib info is requested.")
    arguments = parser.parse_args(args)
    # Configure the logger ASAP
    logger_setup()
    logger = get_logger("bib")
    # Gather requests
    infos: List[Union[Dict, str]] = []
    for f in arguments.files_or_components:
        if os.path.splitext(f)[1].lower() in Extension.yamls:
            infos += [load_input(f)]
        else:  # a single component name, no kind specified
            infos += [f]
    if not infos:
        logger.info("Nothing to do. Pass input files or component names as arguments.")
        return
    print(pretty_repr_bib(*get_bib_info(*infos, logger=logger)))


if __name__ == '__main__':
    bib_script()



from cobaya.likelihoods.base_classes import PlanckClik


class clik(PlanckClik):
    r"""
    Lensing likelihood of Planck's 2018 data release based on temperature+polarization
    map-based lensing reconstruction \cite{Aghanim:2018oex}.
    """
    bibtex_file = 'PlanckLensing2018.bibtex'



import os
import numpy as np
from cobaya.likelihoods.base_classes import InstallableLikelihood
from cobaya.log import LoggedError
from scipy.interpolate import InterpolatedUnivariateSpline


class TT(InstallableLikelihood):
    """
    Python translation of the Planck 2018 Gibbs TT likelihood
    (python Eirik Gjerløw, Feb 2023)
    See https://wiki.cosmos.esa.int/planck-legacy-archive/index.php/CMB_spectrum_%26_Likelihood_Code

    """

    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1",
                       "asset": "planck_2018_lowT.zip",
                       "directory": "planck_2018_lowT_native"}
    lmin: int = 2
    lmax: int = 29
    type = "CMB"
    aliases = ["lowT"]

    @classmethod
    def get_bibtex(cls):
        from cobaya.likelihoods.base_classes import Planck2018Clik
        return Planck2018Clik.get_bibtex()

    def get_requirements(self):
        return {'Cl': {'tt': self.lmax}}

    def get_can_support_params(self):
        return ['A_planck']

    def initialize(self):
        if self.get_install_options() and self.packages_path:
            if self.lmin < 2 or self.lmax > 200 or self.lmin >= self.lmax:
                raise LoggedError(
                    self.log, "lmin must be >= 2, lmax must be <= 200,\n"
                              "and lmin must be less than lmax.")

            path = self.get_path(self.packages_path)

            # The txt files start at l=2, hence the index gymnastics
            cov = np.loadtxt(
                os.path.join(path, 'cov.txt'))[
                  self.lmin - 2:self.lmax + 1 - 2,
                  self.lmin - 2:self.lmax + 1 - 2]
            # The inverse covariance matrix for the gaussian likelihood calculation
            self._covinv = np.linalg.inv(cov)
            # The average cl's for the gaussian likelihood calculation
            self._mu = np.ascontiguousarray(np.loadtxt(
                os.path.join(path, 'mu.txt'))[self.lmin - 2:self.lmax + 1 - 2])
            # The cl's used for offset calculation - hence the full range of ells
            mu_sigma = np.zeros(self.lmax + 1)
            mu_sigma[self.lmin:] = np.loadtxt(
                os.path.join(path, 'mu_sigma.txt'))[self.lmin - 2:self.lmax + 1 - 2]

            # Spline info
            nbins = 1000
            spline_cl = np.loadtxt(
                os.path.join(path, 'cl2x_1.txt'))[:, self.lmin - 2:self.lmax + 1 - 2]
            spline_val = np.loadtxt(
                os.path.join(path, 'cl2x_2.txt'))[:, self.lmin - 2:self.lmax + 1 - 2]
            self._spline = []
            self._spline_derivative = []

            # Set up prior and spline
            self._prior_bounds = np.zeros((self.lmax + 1 - self.lmin, 2))
            for i in range(self.lmax - self.lmin + 1):
                j = 0
                while abs(spline_val[j, i] + 5) < 1e-4:
                    j += 1
                self._prior_bounds[i, 0] = spline_cl[j + 2, i]
                j = nbins - 1
                while abs(spline_val[j, i] - 5) < 1e-4:
                    j -= 1
                self._prior_bounds[i, 1] = spline_cl[j - 2, i]
                self._spline.append(
                    InterpolatedUnivariateSpline(spline_cl[:, i], spline_val[:, i]))
                self._spline_derivative.append(self._spline[-1].derivative())
            # initialize offset to normalize like a chi-squared
            self._offset = 0
            self._offset = self.log_likelihood(mu_sigma)

    def log_likelihood(self, cls_TT, calib=1):
        r"""
        Calculate log likelihood from CMB TT spectrum

        :param cls_TT: L(L+1)C_L/2pi zero-based array in muK^2 units
        :param calib: optional calibration parameter
        :return: log likelihood
        """

        theory: np.ndarray = cls_TT[self.lmin:self.lmax + 1] / calib ** 2

        if any(theory < self._prior_bounds[:, 0]) or any(
                theory > self._prior_bounds[:, 1]):
            return - np.inf

        logl = 0.
        # Convert the cl's to Gaussianized variables
        x = np.zeros_like(theory)
        for i, (spline, diff_spline, cl) in enumerate(
                zip(self._spline, self._spline_derivative, theory)):
            dxdCl = diff_spline(cl)
            if dxdCl < 0:
                return -np.inf
            logl += np.log(dxdCl)
            x[i] = spline(cl)

        delta = x - self._mu
        logl += -0.5 * self._covinv.dot(delta).dot(delta)
        logl -= self._offset

        return logl

    def logp(self, **params_values):
        cls = self.provider.get_Cl(ell_factor=True)['tt']
        return self.log_likelihood(cls, params_values.get('A_planck', 1))



"""
.. module:: cmblikes

:Synopsis: Definition of the CMBlikes class for CMB real or simulated data.
:Author: Antony Lewis (adapted to Cobaya by Jesus Torrado with little modification)

# Load CosmoMC format .dataset files
# AL July 2014 - Dec 2017

"""
# Global
import os
import numpy as np
from typing import List
from getdist import ParamNames, IniFile
from scipy.linalg import sqrtm

# Local
from cobaya.log import LoggedError
from cobaya.likelihoods.base_classes import DataSetLikelihood

CMB_keys = ['tt', 'te', 'ee', 'bb']


class CMBlikes(DataSetLikelihood):
    # Data type for aggregated chi2 (case sensitive)
    type = "CMB"

    # used to form spectra names, e.g. AmapxBmap
    map_separator: str = 'x'

    def get_requirements(self):
        # State requisites to the theory code
        requested_cls = [self.field_names[i] + self.field_names[j]
                         for i, j in zip(*np.where(self.cl_lmax != 0))]
        # l_max has to take into account the window function of the lensing
        # so we check the computed l_max ("l_max" option) is higher than the requested one
        requested_l_max = int(np.max(self.cl_lmax))
        if (getattr(self, "l_max", None) or np.inf) < requested_l_max:
            raise LoggedError(
                self.log, "You are setting a very low l_max. "
                          "The likelihood value will probably not be correct. "
                          "Make sure to make 'l_max'>=%d", requested_l_max)
        self.l_max = max(requested_l_max, getattr(self, "l_max", 0) or 0)
        return {"Cl": {cl: self.l_max for cl in requested_cls}}

    def typeIndex(self, field):
        return self.field_names.index(field)

    def PairStringToMapIndices(self, S):
        if len(S) == 2:
            if self.has_map_names:
                raise LoggedError(self.log, 'CL names must use MAP1xMAP2 names')
            return self.map_names.index(S[0]), self.map_names.index(S[1])
        else:
            try:
                i = S.index(self.map_separator)
            except ValueError:
                raise LoggedError(self.log, 'invalid spectrum name %s' % S)
            return self.map_names.index(S[0:i]), self.map_names.index(S[i + 1:])

    def PairStringToUsedMapIndices(self, used_index, S):
        i1, i2 = self.PairStringToMapIndices(S)
        i1 = used_index[i1]
        i2 = used_index[i2]
        if i2 > i1:
            return i2, i1
        else:
            return i1, i2

    def UseString_to_cols(self, L):
        cl_i_j = self.UseString_to_Cl_i_j(L, self.map_used_index)
        cols = -np.ones(cl_i_j.shape[1], dtype=int)
        for i in range(cl_i_j.shape[1]):
            i1, i2 = cl_i_j[:, i]
            if i1 == -1 or i2 == -1:
                continue
            ix = 0
            for ii in range(self.nmaps):
                for jj in range(ii + 1):
                    if ii == i1 and jj == i2:
                        cols[i] = ix
                    ix += 1
        return cols

    def UseString_to_Cl_i_j(self, S, used_index):
        if not isinstance(S, (list, tuple)):
            S = S.split()
        cl_i_j = np.zeros((2, len(S)), dtype=int)
        for i, p in enumerate(S):
            cl_i_j[:, i] = self.PairStringToUsedMapIndices(used_index, p)
        return cl_i_j

    def MapPair_to_Theory_i_j(self, order, pair):
        i = self.map_fields[order[pair[0]]]
        j = self.map_fields[order[pair[1]]]
        if i <= j:
            return i, j
        else:
            return j, i

    def Cl_used_i_j_name(self, pair):
        return self.Cl_i_j_name(self.used_map_order, pair)

    def Cl_i_j_name(self, names, pair):
        name1 = names[pair[0]]
        name2 = names[pair[1]]
        if self.has_map_names:
            return name1 + self.map_separator + name2
        else:
            return name1 + name2

    def get_cols_from_order(self, order):
        # Converts string Order = TT TE EE XY... or AAAxBBB AAAxCCC BBxCC
        # into indices into array of power spectra (and -1 if not present)
        cols = np.empty(self.ncl, dtype=int)
        cols[:] = -1
        names = order.strip().split()
        ix = 0
        for i in range(self.nmaps):
            for j in range(i + 1):
                name = self.Cl_used_i_j_name([i, j])
                if name not in names and i != j:
                    name = self.Cl_used_i_j_name([j, i])
                if name in names:
                    if cols[ix] != -1:
                        raise LoggedError(self.log,
                                          'get_cols_from_order: duplicate CL type')
                    cols[ix] = names.index(name)
                ix += 1
        return cols

    def elements_to_matrix(self, X, M):
        ix = 0
        for i in range(self.nmaps):
            M[i, 0:i] = X[ix:ix + i]
            M[0:i, i] = X[ix:ix + i]
            ix += i
            M[i, i] = X[ix]
            ix += 1

    def matrix_to_elements(self, M, X):
        ix = 0
        for i in range(self.nmaps):
            X[ix:ix + i + 1] = M[i, 0:i + 1]
            ix += i + 1

    def read_cl_array(self, ini, file_stem, return_full=False):
        # read file of CL or bins (indexed by L)
        filename = ini.relativeFileName(file_stem + '_file')
        cl = np.zeros((self.ncl, self.nbins_used))
        order = ini.string(file_stem + '_order', '')
        if not order:
            incols = last_top_comment(filename)
            if not incols:
                raise LoggedError(self.log, 'No column order given for ' + filename)
        else:
            incols = 'L ' + order
        cols = self.get_cols_from_order(incols)
        data = np.loadtxt(filename)
        Ls = data[:, 0].astype(int)
        if self.binned:
            Ls -= 1
        for i, L in enumerate(Ls):
            if self.bin_min <= L <= self.bin_max:
                for ix in range(self.ncl):
                    if cols[ix] != -1:
                        cl[ix, L - self.bin_min] = data[i, cols[ix]]
        if Ls[-1] < self.bin_max:
            raise LoggedError(
                self.log, 'CMBLikes_ReadClArr: C_l file does not go up to maximum used: '
                          '%s', self.bin_max)
        if return_full:
            return incols.split(), data, cl
        else:
            return cl

    def read_bin_windows(self, ini, file_stem):
        bins = BinWindows(self.pcl_lmin, self.pcl_lmax, self.nbins_used, self.ncl)
        in_cl = ini.split(file_stem + '_in_order')
        out_cl = ini.split(file_stem + '_out_order', in_cl)
        bins.cols_in = self.UseString_to_Cl_i_j(in_cl, self.map_required_index)
        bins.cols_out = self.UseString_to_cols(out_cl)
        norder = bins.cols_in.shape[1]
        if norder != bins.cols_out.shape[0]:
            raise LoggedError(
                self.log, '_in_order and _out_order must have same number of entries')
        bins.binning_matrix = np.zeros(
            (norder, self.nbins_used, self.pcl_lmax - self.pcl_lmin + 1))
        windows = ini.relativeFileName(file_stem + '_files')
        for b in range(self.nbins_used):
            window = np.loadtxt(windows % (b + 1 + self.bin_min))
            err = False
            for i, L in enumerate(window[:, 0].astype(int)):
                if self.pcl_lmin <= L <= self.pcl_lmax:
                    bins.binning_matrix[:, b, L - self.pcl_lmin] = window[i, 1:]
                else:
                    err = err or any(window[i, 1:] != 0)
            if err:
                self.log.warning('%s %u outside pcl_lmin-cl_max range: %s' %
                                 (file_stem, b, windows % (b + 1)))
        if ini.hasKey(file_stem + '_fix_cl_file'):
            raise LoggedError(self.log, 'fix_cl_file not implemented yet')
        return bins

    def init_map_cls(self, nmaps, order):
        if nmaps != len(order):
            raise LoggedError(self.log, 'init_map_cls: size mismatch')

        class CrossPowerSpectrum:
            map_ij: List[int]
            theory_ij: List[int]
            CL: np.ndarray

        cls = np.empty((nmaps, nmaps), dtype=object)
        for i in range(nmaps):
            for j in range(i + 1):
                CL = CrossPowerSpectrum()
                cls[i, j] = CL
                CL.map_ij = [order[i], order[j]]
                CL.theory_ij = self.MapPair_to_Theory_i_j(order, [i, j])
                CL.CL = np.zeros(self.pcl_lmax - self.pcl_lmin + 1)
        return cls

    def init_params(self, ini):
        self.field_names = getattr(self, 'field_names', ['T', 'E', 'B', 'P'])
        self.tot_theory_fields = len(self.field_names)
        self.map_names = ini.split('map_names', default=[])
        self.has_map_names = bool(self.map_names)
        if self.has_map_names:
            # e.g. have multiple frequencies for given field measurement
            map_fields = ini.split('map_fields')
            if len(map_fields) != len(self.map_names):
                raise LoggedError(self.log,
                                  'number of map_fields does not match map_names')
            self.map_fields = [self.typeIndex(f) for f in map_fields]
        else:
            self.map_names = self.field_names
            self.map_fields = np.arange(len(self.map_names), dtype=int)
        fields_use = ini.split('fields_use', [])
        if len(fields_use):
            index_use = [self.typeIndex(f) for f in fields_use]
            use_theory_field = [i in index_use for i in range(self.tot_theory_fields)]
        else:
            if not self.has_map_names:
                raise LoggedError(self.log, 'must have fields_use or map_names')
            use_theory_field = [True] * self.tot_theory_fields
        maps_use = ini.split('maps_use', [])
        if len(maps_use):
            if any(not i for i in use_theory_field):
                self.log.warning('maps_use overrides fields_use')
            self.use_map = [False] * len(self.map_names)
            for j, map_used in enumerate(maps_use):
                if map_used in self.map_names:
                    self.use_map[self.map_names.index(map_used)] = True
                else:
                    raise LoggedError(self.log, 'maps_use item not found - %s' % map_used)
        else:
            self.use_map = [use_theory_field[self.map_fields[i]]
                            for i in range(len(self.map_names))]
        # Bandpowers can depend on more fields than are actually used in likelihood
        # e.g. for correcting leakage or other linear corrections
        self.require_map = self.use_map[:]
        if self.has_map_names:
            if ini.hasKey('fields_required'):
                raise LoggedError(self.log, 'use maps_required not fields_required')
            maps_use = ini.split('maps_required', [])
        else:
            maps_use = ini.split('fields_required', [])
        if len(maps_use):
            for j, map_used in enumerate(maps_use):
                if map_used in self.map_names:
                    self.require_map[self.map_names.index(map_used)] = True
                else:
                    raise LoggedError(self.log, 'required item not found %s' % map_used)
        self.required_theory_field = [False for _ in self.field_names]
        for i in range(len(self.map_names)):
            if self.require_map[i]:
                self.required_theory_field[self.map_fields[i]] = True
        self.ncl_used = 0  # set later reading covmat
        self.like_approx = ini.string('like_approx', 'gaussian')
        self.nmaps = np.count_nonzero(self.use_map)
        self.nmaps_required = np.count_nonzero(self.require_map)
        self.required_order = np.zeros(self.nmaps_required, dtype=int)
        self.map_required_index = -np.ones(len(self.map_names), dtype=int)
        ix = 0
        for i in range(len(self.map_names)):
            if self.require_map[i]:
                self.map_required_index[i] = ix
                self.required_order[ix] = i
                ix += 1
        self.map_used_index = -np.ones(len(self.map_names), dtype=int)
        ix = 0
        self.used_map_order = []
        for i, map_name in enumerate(self.map_names):
            if self.use_map[i]:
                self.map_used_index[i] = ix
                self.used_map_order.append(map_name)
                ix += 1
        self.ncl = (self.nmaps * (self.nmaps + 1)) // 2
        self.pcl_lmax = ini.int('cl_lmax')
        self.pcl_lmin = ini.int('cl_lmin')
        self.binned = ini.bool('binned', True)
        if self.binned:
            self.nbins = ini.int('nbins')
            self.bin_min = ini.int('use_min', 1) - 1
            self.bin_max = ini.int('use_max', self.nbins) - 1
            # needed by read_bin_windows:
            self.nbins_used = self.bin_max - self.bin_min + 1
            self.bins = self.read_bin_windows(ini, 'bin_window')
        else:
            if self.nmaps != self.nmaps_required:
                raise LoggedError(
                    self.log, 'unbinned likelihood must have nmaps==nmaps_required')
            self.nbins = self.pcl_lmax - self.pcl_lmin + 1
            if self.like_approx != 'exact':
                self.log.warning('Unbinned likelihoods untested in this version')
            self.bin_min = ini.int('use_min', self.pcl_lmin)
            self.bin_max = ini.int('use_max', self.pcl_lmax)
            self.nbins_used = self.bin_max - self.bin_min + 1
        self.full_bandpower_headers, self.full_bandpowers, self.bandpowers = \
            self.read_cl_array(ini, 'cl_hat', return_full=True)
        if self.like_approx == 'HL':
            self.cl_fiducial = self.read_cl_array(ini, 'cl_fiducial')
        else:
            self.cl_fiducial = None
        includes_noise = ini.bool('cl_hat_includes_noise', False)
        self.cl_noise = None
        if self.like_approx != 'gaussian' or includes_noise:
            self.cl_noise = self.read_cl_array(ini, 'cl_noise')
            if not includes_noise:
                self.bandpowers += self.cl_noise
            elif self.like_approx == 'gaussian':
                self.bandpowers -= self.cl_noise
        self.cl_lmax = np.zeros((self.tot_theory_fields, self.tot_theory_fields))
        for i in range(self.tot_theory_fields):
            if self.required_theory_field[i]:
                self.cl_lmax[i, i] = self.pcl_lmax
        if self.required_theory_field[0] and self.required_theory_field[1]:
            self.cl_lmax[1, 0] = self.pcl_lmax

        if self.like_approx != 'gaussian':
            cl_fiducial_includes_noise = ini.bool('cl_fiducial_includes_noise', False)
        else:
            cl_fiducial_includes_noise = False
        self.bandpower_matrix = np.zeros((self.nbins_used, self.nmaps, self.nmaps))
        self.noise_matrix = self.bandpower_matrix.copy()
        self.fiducial_sqrt_matrix = self.bandpower_matrix.copy()
        if self.cl_fiducial is not None and not cl_fiducial_includes_noise:
            self.cl_fiducial += self.cl_noise
        for b in range(self.nbins_used):
            self.elements_to_matrix(self.bandpowers[:, b], self.bandpower_matrix[b, :, :])
            if self.cl_noise is not None:
                self.elements_to_matrix(self.cl_noise[:, b], self.noise_matrix[b, :, :])
            if self.cl_fiducial is not None:
                self.elements_to_matrix(self.cl_fiducial[:, b],
                                        self.fiducial_sqrt_matrix[b, :, :])
                self.fiducial_sqrt_matrix[b, :, :] = (
                    sqrtm(self.fiducial_sqrt_matrix[b, :, :]))
        if self.like_approx == 'exact':
            self.fsky = ini.float('fullsky_exact_fksy')
        else:
            self.cov = self.ReadCovmat(ini)
            self.covinv = np.linalg.inv(self.cov)
        if 'linear_correction_fiducial_file' in ini.params:
            self.fid_correction = self.read_cl_array(ini, 'linear_correction_fiducial')
            self.linear_correction = self.read_bin_windows(ini,
                                                           'linear_correction_bin_window')
        else:
            self.linear_correction = None
        if ini.hasKey('nuisance_params'):
            s = ini.relativeFileName('nuisance_params')
            self.nuisance_params = ParamNames(s)
            if ini.hasKey('calibration_param'):
                raise Exception('calibration_param not allowed with nuisance_params')
            if ini.hasKey('calibration_paramname'):
                self.calibration_param = ini.string('calibration_paramname')
            else:
                self.calibration_param = None
        elif ini.string('calibration_param', ''):
            s = ini.relativeFileName('calibration_param')
            if '.paramnames' not in s:
                raise LoggedError(
                    self.log, 'calibration_param must be paramnames file unless '
                              'nuisance_params also specified')
            self.nuisance_params = ParamNames(s)
            self.calibration_param = self.nuisance_params.list()[0]
        else:
            self.calibration_param = None
        if ini.hasKey('log_calibration_prior'):
            self.log.warning('log_calibration_prior in .dataset ignored, '
                             'set separately in .yaml file')
        self.aberration_coeff = ini.float('aberration_coeff', 0.0)

        self.map_cls = self.init_map_cls(self.nmaps_required, self.required_order)

    def ReadCovmat(self, ini):
        """Read the covariance matrix, and the array of which CL are in the covariance,
        which then defines which set of bandpowers are used
        (subject to other restrictions)."""
        covmat_cl = ini.string('covmat_cl', allowEmpty=False)
        self.full_cov = np.loadtxt(ini.relativeFileName('covmat_fiducial'))
        covmat_scale = ini.float('covmat_scale', 1.0)
        cl_in_index = self.UseString_to_cols(covmat_cl)
        self.ncl_used = np.sum(cl_in_index >= 0)
        self.cl_used_index = np.zeros(self.ncl_used, dtype=int)
        cov_cl_used = np.zeros(self.ncl_used, dtype=int)
        ix = 0
        for i, index in enumerate(cl_in_index):
            if index >= 0:
                self.cl_used_index[ix] = index
                cov_cl_used[ix] = i
                ix += 1
        if self.binned:
            num_in = len(cl_in_index)
            pcov = np.empty((self.nbins_used * self.ncl_used,
                             self.nbins_used * self.ncl_used))
            for binx in range(self.nbins_used):
                for biny in range(self.nbins_used):
                    pcov[binx * self.ncl_used: (binx + 1) * self.ncl_used,
                    biny * self.ncl_used: (biny + 1) * self.ncl_used] = (
                            covmat_scale * self.full_cov[
                        np.ix_((binx + self.bin_min) * num_in + cov_cl_used,
                               (biny + self.bin_min) * num_in + cov_cl_used)])
        else:
            raise LoggedError(self.log, 'unbinned covariance not implemented yet')
        return pcov

    # noinspection PyTypeChecker
    def writeData(self, froot):  # pragma: no cover
        np.savetxt(froot + '_cov.dat', self.cov)
        np.savetxt(froot + '_bandpowers.dat', self.full_bandpowers,
                   header=" ".join(self.full_bandpower_headers))
        self.bins.write(froot, 'bin')
        if self.linear_correction is not None:
            self.linear_correction.write(froot, 'linear_correction_bin')

            with open(froot + '_lensing_fiducial_correction', 'w', encoding="utf-8") as f:
                f.write("#%4s %12s \n" % ('bin', 'PP'))
                for b in range(self.nbins):
                    f.write("%5u %12.5e\n" % (b + 1, self.fid_correction[b]))

    def diag_sigma(self):
        return np.sqrt(np.diag(self.full_cov))

    def plot_lensing(self, column='PP', ells=None, units="muK2", ax=None):
        if not np.count_nonzero(self.map_cls):
            raise LoggedError(
                self.log, "No Cl's have been computed yet. "
                          "Make sure you have evaluated the likelihood.")
        try:
            Cl_theo = self.provider.get_Cl(ell_factor=True, units=units)
            Cl = Cl_theo.get(column.lower())
        except KeyError:
            raise LoggedError(self.log, "'%s' spectrum has not been computed." % column)
        import matplotlib.pyplot as plt
        lbin = self.full_bandpowers[:, self.full_bandpower_headers.index('L_av')]
        binned_phicl_err = self.diag_sigma()
        ax = ax or plt.gca()
        bandpowers = self.full_bandpowers[:, self.full_bandpower_headers.index('PP')]
        if 'L_min' in self.full_bandpower_headers:
            lmin = self.full_bandpowers[:, self.full_bandpower_headers.index('L_min')]
            lmax = self.full_bandpowers[:, self.full_bandpower_headers.index('L_max')]
            ax.errorbar(lbin, bandpowers,
                        yerr=binned_phicl_err, xerr=[lbin - lmin, lmax - lbin], fmt='o')
        else:
            ax.errorbar(lbin, bandpowers, yerr=binned_phicl_err, fmt='o')
        if ells is not None:
            Cl = Cl[ells]
        else:
            ells = Cl_theo["ell"]
        ax.plot(ells, Cl, color='k')
        ax.set_xlim([2, ells[-1]])
        return ax

    def get_binned_map_cls(self, Cls, corrections=True):
        band = self.bins.bin(Cls)
        if self.linear_correction is not None and corrections:
            band += self.linear_correction.bin(Cls) - self.fid_correction.T
        return band

    def get_theory_map_cls(self, Cls, data_params=None):
        for i in range(self.nmaps_required):
            for j in range(i + 1):
                CL = self.map_cls[i, j]
                combination = "".join([self.field_names[k] for k in CL.theory_ij]).lower()
                cls = Cls.get(combination)
                if cls is not None:
                    CL.CL[:] = cls[self.pcl_lmin:self.pcl_lmax + 1]
                else:
                    CL.CL[:] = 0
        self.adapt_theory_for_maps(self.map_cls, data_params or {})

    def adapt_theory_for_maps(self, cls, data_params):
        if self.aberration_coeff:
            self.add_aberration(cls)
        self.add_foregrounds(cls, data_params)
        if self.calibration_param is not None and self.calibration_param in data_params:
            for i in range(self.nmaps_required):
                for j in range(i + 1):
                    CL = cls[i, j]
                    if CL is not None:
                        if CL.theory_ij[0] <= 2 and CL.theory_ij[1] <= 2:
                            CL.CL /= data_params[self.calibration_param] ** 2

    def add_foregrounds(self, cls, data_params):
        pass

    def add_aberration(self, cls):
        # adapted from CosmoMC function by Christian Reichardt
        ells = np.arange(self.pcl_lmin, self.pcl_lmax + 1)
        cl_norm = ells * (ells + 1)
        for i in range(self.nmaps_required):
            for j in range(i + 1):
                CL = cls[i, j]
                if CL is not None:
                    if CL.theory_ij[0] <= 2 and CL.theory_ij[1] <= 2:
                        # first get Cl instead of Dl
                        cl_deriv = CL.CL / cl_norm
                        # second take derivative dCl/dl
                        cl_deriv[1:-1] = (cl_deriv[2:] - cl_deriv[:-2]) / 2
                        # handle endpoints approximately
                        cl_deriv[0] = cl_deriv[1]
                        cl_deriv[-1] = cl_deriv[-2]
                        # reapply to Dl's.
                        # note never took 2pi out, so not putting it back either
                        cl_deriv *= cl_norm
                        # also multiply by ell since really wanted ldCl/dl
                        cl_deriv *= ells
                        CL.CL += self.aberration_coeff * cl_deriv

    def write_likelihood_data(self, filename, data_params=None):
        cls = self.init_map_cls(self.nmaps_required, self.required_order)
        self.add_foregrounds(cls, data_params or {})
        with open(filename, 'w', encoding="utf-8") as f:
            cols = []
            for i in range(self.nmaps_required):
                for j in range(i + 1):
                    cols.append(self.Cl_i_j_name(self.map_names, cls[i, j].map_ij))
            f.write('#    L' + ("%17s " * len(cols)) % tuple(cols) + '\n')
            for b in range(self.pcl_lmin, self.pcl_lmax + 1):
                c = [b]
                for i in range(self.nmaps_required):
                    for j in range(i + 1):
                        c.append(cls[i, j].CL[b - self.pcl_lmin])
                f.write(("%I5 " + "%17.8e " * len(cols)) % tuple(c))

    @staticmethod
    def transform(C, Chat, Cfhalf):
        # HL transformation of the matrices
        if C.shape[0] == 1:
            rat = Chat[0, 0] / C[0, 0]
            C[0, 0] = (np.sign(rat - 1) *
                       np.sqrt(2 * np.maximum(0, rat - np.log(rat) - 1)) *
                       Cfhalf[0, 0] ** 2)
            return
        diag, U = np.linalg.eigh(C)
        rot = U.T.dot(Chat).dot(U)
        roots = np.sqrt(diag)
        for i, root in enumerate(roots):
            rot[i, :] /= root
            rot[:, i] /= root
        U.dot(rot.dot(U.T), rot)
        diag, rot = np.linalg.eigh(rot)
        diag = np.sign(diag - 1) * np.sqrt(2 * np.maximum(0, diag - np.log(diag) - 1))
        Cfhalf.dot(rot, U)
        for i, d in enumerate(diag):
            rot[:, i] = U[:, i] * d
        rot.dot(U.T, C)

    def exact_chi_sq(self, C, Chat, L):
        if C.shape[0] == 1:
            return ((2 * L + 1) * self.fsky *
                    (Chat[0, 0] / C[0, 0] - 1 - np.log(Chat[0, 0] / C[0, 0])))
        else:
            M = np.linalg.inv(C).dot(Chat)
            return ((2 * L + 1) * self.fsky *
                    (np.trace(M) - self.nmaps - np.linalg.slogdet(M)[1]))

    def logp(self, **data_params):
        cls = self.provider.get_Cl(ell_factor=True)
        return self.log_likelihood(cls, **data_params)

    # noinspection PyUnboundLocalVariable
    def log_likelihood(self, dls, **data_params):
        r"""
        Get log likelihood from the dls (CMB C_l scaled by L(L+1)/2\pi)

        :param dls: dictionary of d_l ('tt', etc)
        :param data_params: likelihood nuisance parameters
        :return: log likelihood
        """
        self.get_theory_map_cls(dls, data_params)
        C = np.empty((self.nmaps, self.nmaps))
        big_x = np.empty(self.nbins_used * self.ncl_used)
        vecp = np.empty(self.ncl)
        chisq = 0
        if self.binned:
            binned_theory = self.get_binned_map_cls(self.map_cls)
        else:
            Cs = np.zeros((self.nbins_used, self.nmaps, self.nmaps))
            for i in range(self.nmaps):
                for j in range(i + 1):
                    CL = self.map_cls[i, j]
                    if CL is not None:
                        Cs[:, i, j] = CL.CL[self.bin_min - self.pcl_lmin:
                                            self.bin_max - self.pcl_lmin + 1]
                        Cs[:, j, i] = CL.CL[self.bin_min - self.pcl_lmin:
                                            self.bin_max - self.pcl_lmin + 1]
        for b in range(self.nbins_used):
            if self.binned:
                self.elements_to_matrix(binned_theory[b, :], C)
            else:
                C[:, :] = Cs[b, :, :]
            if self.cl_noise is not None:
                C += self.noise_matrix[b]
            if self.like_approx == 'exact':
                chisq += self.exact_chi_sq(
                    C, self.bandpower_matrix[b], self.bin_min + b)
                continue
            elif self.like_approx == 'HL':
                try:
                    self.transform(
                        C, self.bandpower_matrix[b], self.fiducial_sqrt_matrix[b])
                except np.linalg.LinAlgError:
                    self.log.debug("Likelihood computation failed.")
                    return -np.inf
            elif self.like_approx == 'gaussian':
                C -= self.bandpower_matrix[b]
            self.matrix_to_elements(C, vecp)
            big_x[b * self.ncl_used:(b + 1) * self.ncl_used] = vecp[
                self.cl_used_index]
        if self.like_approx == 'exact':
            return -0.5 * chisq
        return -0.5 * self._fast_chi_squared(self.covinv, big_x)


class BinWindows:
    cols_in: np.ndarray
    cols_out: np.ndarray
    binning_matrix: np.ndarray

    def __init__(self, lmin, lmax, nbins, ncl):
        self.lmin = lmin
        self.lmax = lmax
        self.nbins = nbins
        self.ncl = ncl

    def bin(self, theory_cl, cls=None):
        if cls is None:
            cls = np.zeros((self.nbins, self.ncl))
        for i, ((x, y), ix_out) in enumerate(zip(self.cols_in.T, self.cols_out)):
            cl = theory_cl[x, y]
            if cl is not None and ix_out >= 0:
                cls[:, ix_out] += np.dot(self.binning_matrix[i, :, :], cl.CL)
        return cls

    def write(self, froot, stem):
        if not os.path.exists(froot + stem + '_window'):
            os.mkdir(froot + '_window')
        for b in range(self.nbins):
            with open(froot + stem + '_window/window%u.dat' % (b + 1),
                      'w', encoding="utf-8") as f:
                for L in np.arange(self.lmin[b], self.lmax[b] + 1):
                    f.write(
                        ("%5u " + "%10e" * len(self.cols_in) + "\n") %
                        (L, self.binning_matrix[:, b, L]))


def last_top_comment(fname):
    result = None
    with open(fname, encoding="utf-8-sig") as f:
        while x := f.readline():
            if x := x.strip():
                if x[0] != '#':
                    return result
                result = x[1:].strip()
    return None


def white_noise_from_muK_arcmin(noise_muK_arcmin):
    return (noise_muK_arcmin * np.pi / 180 / 60.) ** 2


def save_cl_dict(filename, array_dict, lmin=2, lmax=None,
                 cl_dict_lmin=0):  # pragma: no cover
    """
    Save a Cobaya dict of CL to a text file, with each line starting with L.

    :param filename: filename to save
    :param array_dict: dictionary of power spectra
    :param lmin: minimum L to save
    :param lmax: maximum L to save
    :param cl_dict_lmin: L to start output in file (usually 0 or 2)
    """
    cols = []
    labels = []
    for key in CMB_keys:
        if key in array_dict:
            lmax = lmax or array_dict[key].shape[0] - 1 + cl_dict_lmin
            cols.append(array_dict[key][lmin - cl_dict_lmin:lmax - cl_dict_lmin + 1])
            labels.append(key.upper())
    if 'pp' in array_dict:
        lmax = lmax or array_dict['pp'].shape[0] - 1 + cl_dict_lmin
        cols.append(array_dict['pp'][lmin - cl_dict_lmin:lmax - cl_dict_lmin + 1])
        labels.append('PP')
    ls = np.arange(lmin, lmax + 1)
    np.savetxt(filename, np.vstack((ls,) + tuple(cols)).T,
               fmt=['%4u'] + ['%12.7e'] * len(cols),
               header=' L ' + ' '.join(['{:13s}'.format(lab) for lab in labels]))


def make_forecast_cmb_dataset(fiducial_Cl, output_root, output_dir=None,
                              noise_muK_arcmin_T=None,
                              noise_muK_arcmin_P=None,
                              NoiseVar=None, ENoiseFac=2, fwhm_arcmin=None,
                              lmin=2, lmax=None, fsky=1.0,
                              lens_recon_noise=None, cl_dict_lmin=0):  # pragma: no cover
    """
    Make a simulated .dataset and associated files with 'data' set at the input fiducial
    model. Uses the exact full-sky log-likelihood, scaled by fsky.

    If you want to use numerical N_L CMB noise files, you can just replace the noise
    .dat text file produced by this function.

    :param fiducial_Cl: dictionary of Cls to use, combination of tt, te, ee, bb, pp;
                        note te must be included with tt and ee when using them
    :param output_root: root name for output files, e.g. 'my_sim1'
    :param output_dir: output directory
    :param noise_muK_arcmin_T: temperature noise in muK-arcmin
    :param noise_muK_arcmin_P: polarization noise in muK-arcmin
    :param NoiseVar: alternatively if noise_muK_arcmin_T is None, effective
        isotropic noise variance for the temperature (N_L=NoiseVar with no beam)
    :param ENoiseFac: factor by which polarization noise variance is higher than
                NoiseVar (usually 2, for Planck about 4
                        as only half the detectors polarized)
    :param fwhm_arcmin: beam fwhm in arcminutes
    :param lmin: l_min
    :param lmax: l_max
    :param fsky: sky fraction
    :param lens_recon_noise: optional array, starting at L=0, for the
       pp lensing reconstruction noise, in [L(L+1)]^2C_L^phi/2pi units
    :param cl_dict_lmin: l_min for the arrays in fiducial_Cl
    :return: IniFile that was saved
    """
    ini = IniFile()
    dataset = ini.params

    cl_keys = fiducial_Cl.keys()
    use_CMB = set(cl_keys).intersection(CMB_keys)
    use_lensing = lens_recon_noise is not None

    if use_CMB:
        if NoiseVar is None:
            if noise_muK_arcmin_T is None:
                raise ValueError('Must specify noise')
            NoiseVar = white_noise_from_muK_arcmin(noise_muK_arcmin_T)
            if noise_muK_arcmin_P is not None:
                ENoiseFac = (noise_muK_arcmin_P / noise_muK_arcmin_T) ** 2
        elif noise_muK_arcmin_T is not None or noise_muK_arcmin_P is not None:
            raise ValueError('Specific either noise_muK_arcmin or NoiseVar')
        fields_use = ''
        if 'tt' in cl_keys or 'te' in cl_keys:
            fields_use = 'T'
        if 'ee' in cl_keys or 'te' in cl_keys:
            fields_use += ' E'
        if 'bb' in cl_keys:
            fields_use += ' B'
        if 'pp' in cl_keys and use_lensing:
            fields_use += ' P'
        if 'tt' in cl_keys and 'ee' in cl_keys and 'te' not in cl_keys:
            raise ValueError('Input power spectra should have te if using tt and ee -'
                             'using the exact likelihood requires the full covariance.')
    else:
        fields_use = 'P'

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    dataset['fields_use'] = fields_use

    if use_CMB:
        fwhm = fwhm_arcmin / 60
        xlc = 180 * np.sqrt(8. * np.log(2.)) / np.pi
        sigma2 = (fwhm / xlc) ** 2
        noise_cols = 'TT           EE          BB'
        if use_lensing:
            noise_cols += '          PP'
    elif use_lensing:
        noise_cols = 'PP'
    else:
        raise ValueError('Must use CMB or lensing C_L')
    noise_file = output_root + '_Noise.dat'
    with open(os.path.join(output_dir, noise_file), 'w') as f:
        f.write('#L %s\n' % noise_cols)

        for ell in range(lmin, lmax + 1):
            noises = []
            if use_CMB:
                # noinspection PyUnboundLocalVariable
                noise_cl = ell * (ell + 1.) / 2 / np.pi * NoiseVar * np.exp(
                    ell * (ell + 1) * sigma2)
                noises += [noise_cl, ENoiseFac * noise_cl, ENoiseFac * noise_cl]
            if use_lensing:
                noises += [lens_recon_noise[ell]]
            f.write("%d " % ell + " ".join("%E" % elem for elem in noises) + "\n")

    dataset['fullsky_exact_fksy'] = fsky
    dataset['dataset_format'] = 'CMBLike2'
    dataset['like_approx'] = 'exact'

    dataset['cl_lmin'] = lmin
    dataset['cl_lmax'] = lmax

    dataset['binned'] = False

    dataset['cl_hat_includes_noise'] = False

    save_cl_dict(os.path.join(output_dir, output_root + '.dat'),
                 fiducial_Cl, cl_dict_lmin=cl_dict_lmin)
    dataset['cl_hat_file'] = output_root + '.dat'
    dataset['cl_noise_file '] = noise_file

    ini.saveFile(os.path.join(output_dir, output_root + '.dataset'))
    return ini



r"""
.. module:: prior

:Synopsis: Class containing the prior and reference pdf, and other parameter information.
:Author: Jesus Torrado


Basic parameter specification
-----------------------------

The ``params`` block contains all the information intrinsic
to the parameters of the model: their prior pdf and reference pdf, their latex labels,
and useful properties for particular samplers (e.g. width of a proposal pdf in an MCMC
-- see the documentation for each sampler).
The prior and reference pdf are managed by the :class:`Prior` class.

You can specify three different kinds of parameters:

+ **Fixed** parameters are specified by assigning them a value, and are passed directly to
  the likelihood or theory code.
+ **Sampled** parameters need a ``prior`` pdf definition and, optionally,
  a reference pdf (``ref``),
  a LaTeX label to be used when producing plots,
  and additional properties to aid particular samplers.
+ **Derived** parameters do **not** have a prior definition
  (neither a reference value or pdf),
  but can have a LaTeX label and a ``min`` and/or ``max``
  to be used in the sample analysis
  (e.g. to guess the pdf tails correctly if the derived quantity needs to be positive or
  negative -- defaulted to ``-.inf``, ``.inf`` resp.).

The (optional) **reference** pdf (``ref``) for **sampled** parameters defines the region
of the prior which is of most interest (e.g. where most of the prior mass is expected);
samplers needing an initial point for a chain will attempt to draw it from the ``ref``
if it has been defined (otherwise from the prior). A good reference pdf will avoid a long
*burn-in* stage during the sampling. If you assign a single value to ``ref``, samplers
will always start from that value; however this makes convergence tests less reliable as
each chain will start from the same point (so all chains could be stuck near the same
point).

.. note::

   When running in parallel with MPI you can give ``ref`` different values for each of
   the different parallel processes (either at initialisation or by calling
   :func:`Prior.set_reference`). They will be taken into account e.g. for starting
   parallel MCMC chains at different points/distributions of your choice.

The syntax for priors and ref's has the following fields:

+ ``dist`` (default: ``uniform``): any 1-dimensional continuous distribution from
  `scipy.stats <https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions>`_;
  e.g. ``uniform``, ``[log]norm`` for a [log-]gaussian,
  or ``halfnorm`` for a half-gaussian.
+ ``loc`` and ``scale`` (default: 0 and 1, resp.): the *location* and *scale* of the pdf,
  as they are understood for each particular pdf in :class:`scipy.stats`; e.g. for a
  ``uniform`` pdf, ``loc`` is the lower bound and ``scale`` is the length of the domain,
  whereas in a gaussian (``norm``) ``loc`` is the mean and ``scale`` is the standard
  deviation.
+ Additional specific parameters of the distribution, e.g. ``a`` and ``b`` as the powers
  of a Beta pdf.

.. note::

   For bound distributions (e.g. ``uniform``, ``beta``...), you can use the more
   intuitive arguments ``min`` and ``max`` (default: 0 and 1 resp.) instead of ``loc`` and
   ``scale`` (NB: unexpected behaviour for an unbounded pdf).


The order of the parameters is conserved in the table of samples, except that
derived parameters are always moved to the end.

An example ``params`` block:

.. code-block :: yaml

     params:

       A: 1 # fixed!

       B1:  # sampled! with uniform prior on [-1,1] and ref pdf N(mu=0,sigma=0.25)
         prior:
           min: -1
           max:  1
         ref:
           dist: norm
           loc:   0
           scale: 0.25
         latex: \mathcal{B}_2
         proposal: 0.25

       B2:  # sampled! with prior N(mu=0.5,sigma=3) and fixed reference value 1
         prior:
           dist: norm
           loc: 0.5
           scale: 3
         ref: 1  # fixed reference value: all chains start here!
         latex: \mathcal{B}_2
         proposal: 0.5

       C:  # derived!
         min: 0
         latex: \mathcal{C}

You can find another basic example :ref:`here <example_quickstart_shell>`.


.. _prior_external:

Multidimensional priors
-----------------------

The priors described above are obviously 1-dimensional.
You can also define more complicated, multidimensional priors
using a ``prior`` block at the base level (i.e. not indented), as in the
:ref:`the advanced example <example_advanced_shell>`. All the details below also apply to
the definition of :ref:`external likelihoods <likelihood_external>`. We will call these
custom priors "`external` priors".

Inside the ``prior`` block, list a pair of priors as ``[name]: [function]``, where the
functions must return **log**-priors. These priors will be multiplied by the
one-dimensional ones defined above. Even if you define a prior for some parameters
in the ``prior`` block, you still have to specify their bounds in the ``params`` block.

A prior function can be specified in two different ways:

a) **As a function**, either assigning a function defined with ``def``, or a ``lambda``.
   Its arguments must be known parameter names.

b) **As a string,** which will be passed to ``eval()``. The string can be a
   ``lambda`` function definition with known parameters as its arguments (you can
   use ``scipy.stats`` and  ``numpy`` as ``stats`` and ``np`` resp.), or an
   ``import_module('filename').function`` statement if your function is defined in a
   separate file ``filename`` `in the current folder` and is named ``function``
   (see e.g. :ref:`the advanced example <example_advanced_shell>`).

.. note::

   Your function definition may contain keyword arguments that aren't known parameter
   names. They will simply be ignored and left to their default.

.. warning::

   When **resuming** a run using an **external** python object as input (e.g. a prior or a
   likelihood), there is no way for Cobaya to know whether the new object is the same as
   the old one: it is left to the user to ensure reproducibility of those objects between
   runs.

.. warning::

   External priors can only be functions **sampled** and **fixed**
   and **derived** parameters that are dynamically defined in terms of other inputs.
   Derived parameters computed by the theory code cannot be used in a prior, since
   otherwise the full prior could not be computed **before** the likelihood,
   preventing us from avoiding computing the likelihood when the prior is null, or
   forcing a *post-call* to the prior.

   **Workaround:** Define your function as a
   :ref:`external likelihood <likelihood_external>` instead, since likelihoods do have
   access to derived parameters (for an example, see `this comment
   <https://github.com/CobayaSampler/cobaya/issues/18#issuecomment-447818811>`_).


Prior normalization for evidence computation
--------------------------------------------

The one-dimensional priors defined within the ``params`` block are automatically
normalized, so any sampler that computes the evidence will produce the right results as
long as no external priors have been defined, whose normalization is unknown.

To get the prior normalization if using external functions as priors, you can substitute
your likelihood by the :doc:`dummy unit likelihood <likelihood_one>`, and make an initial
run with :doc:`PolyChord <sampler_polychord>` to get the prior volume
(see section :ref:`polychord_bayes_ratios`).

In general, avoid improper priors, since they will produce unexpected errors, e.g.

.. code-block:: yaml

    params:
      a:
        prior:
          min: -.inf
          max:  .inf


.. _repar:

Defining parameters dynamically
-------------------------------

We may want to sample in a parameter space different from the one understood by the
likelihood, e.g. because we expect the posterior to be simpler on the alternative
parameters.

For instance, in the :doc:`advanced example <example_advanced>`, the posterior on the
radius and the angle is a gaussian times a uniform, instead of a more complicated
gaussian ring. This is done in a simple way at
:ref:`the end of the example <example_advanced_rtheta>`.
Let us discuss the general case here.

To enable this, **cobaya** creates a `re-parameterization` layer between the `sampled`
parameters, and the `input` parameters of the likelihood. E.g. if we want to **sample**
from the logarithm of an **input** parameter of the likelihood, we would do:

.. code:: yaml

    params:
      logx:
        prior: [...]
        drop: True
      x: "lambda logx: np.exp(logx)"
      y:
        derived: "lambda x: x**2"

When that information is loaded **cobaya** will create an interface between two sets of
parameters:

+ **Sampled** parameters, i.e. those that have a prior, including ``logx`` but not ``x``
  (it does not have a prior: it's fixed by a function). Those are the ones with which the
  **sampler** interacts. Since the likelihood does not understand them, they must be
  **dropped** with ``drop: True`` if you are using any parameter-agnostic components.

+ Likelihood **input** parameters: those that will be passed to the likelihood
  (or theory). They are identified by either having a prior and not being **dropped**, or
  having being assigned a fixed value or a function. Here, ``x`` would be an input
  parameter, but not ``logx``.

.. _cobaya_diagram:
.. image:: ./img/diagram.svg
   :align: center
   :width: 60%

We can use a similar approach to define dynamical **derived** parameters, which can depend
on *input* and *sampled* parameters. To distinguish their notation from that of input
parameters, we insert the functions defining them under a ``derived`` property
(see the parameter ``y`` in the example above).

.. note::

   **Dynamical derived** parameters can also be functions of yet-undefined parameters.
   In that case, those parameters will be automatically requested from the likelihood (or
   theory code) that understands them.

.. note::

   Also available are the :math:`\chi^2 = -2 \log \mathcal{L}` of the used likelihoods,
   as `chi2__[name]`, e.g.

   .. code:: yaml

     likelihood:
       my_like: [...]
     params:
       my_derived:
         derived: "lambda chi2__my_like: [...]"

.. note::

   By default, the values of **dynamical input** parameters (e.g. ``x`` above) are saved
   as if they were derived parameters. If you would like to ignore them, define them using
   the following *extended notation*:

   .. code:: yaml

      params:
        x:
          value: "lambda logx: np.exp(logx)"
          derived: False

.. note::

   If you want to fix the value of a parameter whose only role is being an argument of a
   dynamically defined one and is *not supposed to be passed to the likelihood*, you need
   to explicitly *drop* it. E.g. suppose that you want to sample from a likelihood that
   depends on ``x``, but want to use ``log(x)`` as the sampled parameter; you would do it
   like this:

   .. code:: yaml

      params:
        logx:
          prior: [...]
          drop: True
        x:
          value: "lambda logx: np.exp(x)"

   Now, if you want to fix the value of ``logx`` without changing the structure of the
   input, do

   .. code:: yaml

      params:
        logx:
          value: [fixed_value]
          drop: True
        x:
          value: "lambda logx: np.exp(x)"


.. _prior_inheritance:

Changing and redefining parameters; inheritance
-----------------------------------------------

As we will see in the :doc:`likelihoods documentation <likelihoods>`, just by mentioning a
likelihood, the parameters of its default experimental model are
inherited without needing to re-state them.

But you may want to fix their value, re-define their prior or some other property in
your input file. To do that, simply state you new parameter definition as you would
normally do. For your convenience, if you do that, you don't need to re-state all of the
properties of the parameters (e.g. the latex label if just changing the prior).

As a general rule, when trying to redefine anything in a parameter, everything not
re-defined is inherited, except for the prior, which must not be inheritable in order to
allow re-defining sampled parameters into something else.

Just give it a try and it should work fine, but, in case you need the details:

* Re-defining **fixed** into:

 - **fixed**: simply assign it a new value
 - **sampled**: define a prior for it (and optionally reference, label, etc.)
 - **derived**: mention the parameter and assign nothing to it (you may define a
   min, max or latex label, but *not a prior*)

* Re-defining **sampled** into:

 - **fixed**: simply assign it a value
 - **sampled**: change any of prior/ref/latex, etc, to your taste; the rest are inherited
   from the defaults
 - **derived**: mention the parameter and assign nothing to it (you may define a
   min, max or latex label, but *not a prior*; the label is inherited from the defaults)

* Re-defining **derived** into:

 - **fixed**: simply assign it a value
 - **sampled**: define a prior for it (and optionally reference, label, etc.)
 - **derived**: change any of prior/ref/latex, etc, to your taste; the rest are inherited
   from the defaults

"""

# Global
import numbers
from types import MethodType
from typing import Sequence, NamedTuple, Callable, Optional, Mapping, List
import numpy as np

# Local
from cobaya.conventions import prior_1d_name
from cobaya.typing import PriorsDict
from cobaya.tools import get_external_function, get_scipy_1d_pdf, read_dnumber
from cobaya.tools import _fast_norm_logpdf, getfullargspec
from cobaya.log import LoggedError, HasLogger
from cobaya.parameterization import Parameterization

# Fast logpdf for uniforms and norms (do not understand nan masks!)
fast_logpdfs = {"norm": _fast_norm_logpdf}


class ExternalPrior(NamedTuple):
    logp: Callable
    params: Sequence[str]


class Prior(HasLogger):
    """
    Class managing the prior and reference pdf's.
    """

    def __init__(self, parameterization: Parameterization,
                 info_prior: Optional[PriorsDict] = None):
        """
        Initializes the prior and reference pdf's from the input information.
        """
        self.set_logger()
        self._parameterization = parameterization
        sampled_params_info = parameterization.sampled_params_info()
        # pdf: a list of independent components
        # in principle, separable: one per parameter
        self.params = []
        self.pdf = []
        self._bounds = np.zeros((len(sampled_params_info), 2))
        for i, p in enumerate(sampled_params_info):
            self.params += [p]
            prior = sampled_params_info[p].get("prior")
            self.pdf += [get_scipy_1d_pdf({p: prior})]
            fast_logpdf = fast_logpdfs.get(self.pdf[-1].dist.name)
            if fast_logpdf:
                self.pdf[-1].logpdf = MethodType(fast_logpdf, self.pdf[-1])
            self._bounds[i] = [-np.inf, np.inf]
            try:
                self._bounds[i] = self.pdf[-1].interval(1)
            except AttributeError as excpt:
                raise LoggedError(
                    self.log,
                    "No bounds defined for parameter '%s' "
                    "(maybe not a scipy 1d pdf).",
                    p
                ) from excpt
        self._uniform_indices = np.array(
            [i for i, pdf in enumerate(self.pdf) if pdf.dist.name == 'uniform'],
            dtype=int)
        self._non_uniform_indices = np.array(
            [i for i in range(len(self.pdf)) if i not in self._uniform_indices],
            dtype=int)
        self._non_uniform_logpdf = [self.pdf[i].logpdf for i in self._non_uniform_indices]
        self._upper_limits = self._bounds[:, 1].copy()
        self._lower_limits = self._bounds[:, 0].copy()
        self._uniform_logp = -np.sum(np.log(self._upper_limits[self._uniform_indices] -
                                            self._lower_limits[self._uniform_indices]))
        # Set the reference pdf's
        self._ref_is_pointlike: Optional[bool] = None
        self.set_reference({p: v.get("ref") for p, v in sampled_params_info.items()})
        # Process the external prior(s):
        self.external = {}
        self.external_dependence = set()
        info_prior = info_prior or {}
        for name in info_prior:
            if name == prior_1d_name:
                raise LoggedError(self.log, "The name '%s' is a reserved prior name. "
                                            "Please use a different one.", prior_1d_name)
            self.log.debug(
                "Loading external prior '%s' from: '%s'", name, info_prior[name])
            logp = get_external_function(info_prior[name], name=name)
            argspec = getfullargspec(logp)
            known = set(parameterization.input_params())
            params = [p for p in argspec.args if p in known]
            params_without_default = set(
                argspec.args[:(len(argspec.args) - len(argspec.defaults or []))])
            unknown = params_without_default - known
            if unknown:
                if unknown.intersection(parameterization.derived_params()):
                    err = ("External prior '%s' has arguments %s that are output derived "
                           "parameters, Priors must be functions of input parameters. "
                           "Use a separate 'likelihood' for the prior if needed.")
                else:
                    err = ("Some arguments of the external prior '%s' cannot be "
                           "found and don't have a default value either: %s")
                raise LoggedError(self.log, err, name, list(unknown))
            self.external_dependence.update(params)
            self.external[name] = ExternalPrior(logp=logp, params=params)
            self.mpi_warning("External prior '%s' loaded. "
                             "Mind that it might not be normalized!", name)

        parameterization.check_dropped(self.external_dependence)

    def d(self):
        """
        Returns:
           Dimensionality of the parameter space.
        """
        return len(self.pdf)

    # Not very useful, except for getting the prior names as list(self)
    # Created for consistency with likelihoods
    def __iter__(self):
        return (p for p in [prior_1d_name] + list(self.external))

    def __len__(self):
        return 1 + len(self.external)

    def bounds(
            self, confidence: float = 1, confidence_for_unbounded: float = 1
    ) -> np.ndarray:
        """
        Returns a list of bounds ``[min, max]`` per parameter, containing confidence
        intervals of a certain ``confidence`` level, centered around the median, by
        default (``confidence=1`` the full parameter range).

        For unbounded parameters, if ``confidence=1``, one can specify some value slightly
        smaller than 1 for ``confidence_for_unbounded``, in order to ensure that all
        bounds returned are finite.

        NB: If an external prior has been defined, the bounds given in the 'prior'
        sub-block of that particular parameter's info may not be faithful to the
        externally defined prior. A warning will be raised in that case.

        Parameters
        ----------
        confidence : float, default 1
            Probability mass contained within the returned bounds. Capped at 1.

        confidence_for_unbounded : float, default 1
            Confidence level applied to the unbounded parameters if ``confidence=1``;
            ignored otherwise.

        Returns
        -------
        bounds : 2-d array [[param1_min, param1_max], ...]
            Array of bounds ``[min,max]`` for the parameters, in the order given by the
            input.

        Raises
        ------
        LoggedError
            If some parameters do not have bounds defined.
        """
        if confidence < 1:
            return np.array([pdf.interval(confidence) for pdf in self.pdf])
        # Else, confidence >= 1:
        if confidence_for_unbounded >= 1:
            return self._bounds
        else:
            bounds = self._bounds.copy()
            infs = list(set(np.argwhere(np.isinf(bounds)).T[0]))
            try:
                if infs:
                    self.mpi_warning(
                        "There are unbounded parameters (%r). Prior bounds "
                        "are given at %s confidence level. Beware of "
                        "likelihood modes at the edge of the prior",
                        [self.params[ix] for ix in infs],
                        confidence_for_unbounded
                    )
                    bounds[infs] = [
                        self.pdf[i].interval(confidence_for_unbounded) for i in infs]
                return bounds
            except AttributeError as excpt:
                raise LoggedError(
                    self.log,
                    "Some parameter names (positions %r) have no bounds defined.",
                    infs
                ) from excpt

    def sample(self, n=1, ignore_external=False, random_state=None):
        """
        Generates samples from the prior pdf.

        If an external prior has been defined, it is not possible to sample from the prior
        directly. In that case, if you want to sample from the "default" pdf (i.e.
        ignoring the external prior), set ``ignore_external=True``.

        Returns:
          Array of ``n`` samples from the prior, as vectors ``[value of param 1, ...]``.
        """
        if not ignore_external and self.external:
            raise LoggedError(
                self.log, "It is not possible to sample from an external prior "
                          "(see help of this function on how to fix this).")
        return np.array([pdf.rvs(n, random_state=random_state) for pdf in self.pdf]).T

    def logps(self, x: np.ndarray) -> List[float]:
        """
        Takes a point (sampled parameter values, in the correct order).

        Returns:
           An array of the prior log-probability densities of the given point
           or array of points. The first element on the list is the products
           of 1d priors specified in the ``params`` block, and the following
           ones (if present) are the priors specified in the ``prior`` block
           in the same order.
        """
        logps = self.logps_internal(x)
        if logps != -np.inf:
            if self.external:
                input_params = self._parameterization.to_input(x)
                return [logps] + self.logps_external(input_params)
            else:
                return [logps]
        else:
            return [-np.inf] * (1 + len(self.external))

    def logp(self, x: np.ndarray):
        """
        Takes a point (sampled parameter values, in the correct order).

        Returns:
           The prior log-probability density of the given point or array of points.
        """
        return np.sum(self.logps(x), axis=0)

    def logps_internal(self, x: np.ndarray) -> float:
        """
        Takes a point (sampled parameter values, in the correct order).

        Returns:
           The prior log-probability density of the given point
           or array of points, only including the products
           of 1d priors specified in the ``params`` block, no external priors
        """
        self.log.debug("Evaluating prior at %r", x)
        if all(x <= self._upper_limits) and all(x >= self._lower_limits):
            # Apparently faster to sum list than generator (for short enough lists)
            logps = self._uniform_logp + (sum([logpdf(xi) for logpdf, xi in
                                               zip(self._non_uniform_logpdf,
                                                   x[self._non_uniform_indices])])
                                          if len(self._non_uniform_indices) else 0)
        else:
            logps = -np.inf
        self.log.debug("Got logpriors (internal) = %r", logps)
        return logps

    def logps_external(self, input_params) -> List[float]:
        """Evaluates the logprior using the external prior only."""
        logps = [ext.logp(**{p: input_params[p] for p in ext.params})
                 for ext in self.external.values()]
        self.log.debug("Got logpriors (external) = %r", logps)
        return logps

    def covmat(self, ignore_external=False):
        """
        Returns:
           The covariance matrix of the prior.
        """
        if not ignore_external and self.external:
            raise LoggedError(
                self.log,
                "It is not possible to get the covariance matrix from an external prior.")
        return np.diag([pdf.var() for pdf in self.pdf]).T

    def set_reference(self, ref_info):
        """
        Sets or updates the reference pdf with the given parameter input info.

        ``ref_info`` should be a dict ``{parameter_name: [ref definition]}``, not
        ``{parameter_name: {"ref": [ref definition]}}``.

        When called after prior initialisation, not mentioning a parameter leaves
        its reference pdf unchanged, whereas explicitly setting ``ref: None`` sets
        the prior as the reference pdf.

        You can set different reference pdf's/values for different MPI processes,
        e.g. for fixing different starting points for parallel MCMC chains.
        """
        if not hasattr(self, "ref_pdf"):
            # Initialised with nan's in case ref==None: no ref -> uses prior
            self.ref_pdf = [np.nan] * self.d()
        unknown = set(ref_info).difference(self.params)
        if unknown:
            raise LoggedError(self.log,
                              f"Cannot set reference pdf for parameter(s) {unknown}: "
                              "not sampled parameters.")
        for i, p in enumerate(self.params):
            # The next if ensures correct behavious in "update call",
            # where not mentioning a parameter and making its ref None are different
            # (not changing vs setting to prior)
            if p not in ref_info:
                continue  # init: use prior; update: don't change
            ref = ref_info[p]
            # [number, number] interpreted as Gaussian
            if isinstance(ref, Sequence) and len(ref) == 2 and all(
                    isinstance(n, numbers.Number) for n in ref):
                ref = {"dist": "norm", "loc": ref[0], "scale": ref[1]}
            if isinstance(ref, numbers.Real):
                self.ref_pdf[i] = float(ref)
            elif isinstance(ref, Mapping):
                self.ref_pdf[i] = get_scipy_1d_pdf({p: ref})
            elif ref is None:
                # We only get here if explicit `param: None` mention!
                self.ref_pdf[i] = np.nan
            else:
                raise LoggedError(self.log,
                                  "'ref' for starting position should be None or a number"
                                  ", a list of two numbers for normal mean and deviation,"
                                  "or a dict with parameters for a scipy distribution.")
        # Re-set the pointlike-ref property
        self._set_pointlike()

    @property
    def reference_is_pointlike(self) -> bool:
        """
        Whether there is a fixed reference point for all parameters, such that calls to
        :func:`Prior.reference` would always return the same.
        """
        if not hasattr(self, "_ref_is_pointlike"):
            self._set_pointlike()
        return self._ref_is_pointlike

    def _set_pointlike(self):
        self._ref_is_pointlike = all(
            # np.nan is a numbers.Number instance, but not a fixed ref (uses prior)
            (isinstance(ref, numbers.Number) and not np.isnan(ref))
            for ref in self.ref_pdf)

    def reference(self, max_tries=np.inf, warn_if_tries="10d", ignore_fixed=False,
                  warn_if_no_ref=True, random_state=None) -> np.ndarray:
        """
        Returns:
          One sample from the ref pdf. For those parameters that do not have a ref pdf
          defined, their value is sampled from the prior.

        If `ignored_fixed=True` (default: `False`), fixed reference values will be ignored
        in favor of the full prior, ensuring some randomness for all parameters (useful
        e.g. to prevent caching when measuring speeds).

        NB: The way this function works may be a little dangerous:
        if two parameters have an (external)
        joint prior defined and only one of them has a reference pdf, one should
        sample the other from the joint prior *conditioned* to the first one being
        drawn from the reference. Otherwise, one may end up with a point with null
        prior pdf. In any case, it should work for getting initial reference points,
        e.g. for an MCMC chain.
        """
        if np.nan in self.ref_pdf and warn_if_no_ref:
            self.log.info(
                "Reference values or pdfs for some parameters were not provided. "
                "Sampling from the prior instead for those parameters.")
        # As a curiosity, `r is np.nan` was returning False after `r = np.nan` if
        # it had been passed via MPI before the test, since this creates a "new" np.nan
        # NB: isinstance(np.nan, numers.Real) --> True
        where_ignore_ref = [
            isinstance(r, numbers.Real) and (np.isnan(r) or ignore_fixed)
            for r in self.ref_pdf
        ]
        tries = 0
        warn_if_tries = read_dnumber(warn_if_tries, self.d())
        ref_sample = np.empty(len(self.ref_pdf))
        while tries < max_tries:
            tries += 1
            if any(where_ignore_ref):
                prior_sample = self.sample(ignore_external=True,
                                           random_state=random_state)[0]
                ref_sample[where_ignore_ref] = prior_sample[where_ignore_ref]
            for i, ref_pdf in enumerate(self.ref_pdf):
                if not where_ignore_ref[i]:
                    if hasattr(ref_pdf, "rvs"):
                        ref_sample[i] = ref_pdf.rvs(
                            random_state=random_state)  # type: ignore
                    else:
                        ref_sample[i] = ref_pdf.real

            if self.logp(ref_sample) > -np.inf:
                return ref_sample
            if tries == warn_if_tries:
                self.log.warning(
                    "If stuck here, maybe it is not possible to sample from the "
                    "reference pdf a point with non-null prior. Check that the reference "
                    "pdf and the prior are consistent.")
        if self.reference_is_pointlike:
            raise LoggedError(self.log, "The reference point provided has null prior. "
                                        "Set 'ref' to a different point or a pdf.")
        raise LoggedError(
            self.log, "Could not sample from the reference pdf a point with non-"
                      "null prior density after %d tries. "
                      "Maybe your prior is improper of your reference pdf is "
                      "null-defined in the domain of the prior.", max_tries)

    def reference_variances(self):
        """
        Returns:
          The standard variances of the 1d ref pdf's. For those parameters that do not
          have a proposal defined, the standard deviation of the prior can be taken
          as a starting point to estimate one.
        """
        variances = np.array([getattr(ref_pdf, "var", lambda: np.nan)()
                              for i, ref_pdf in enumerate(self.ref_pdf)])
        where_no_ref = np.isnan(variances)
        if np.any(where_no_ref):
            self.mpi_warning("Reference pdf not defined or improper for some parameters. "
                             "Using prior's sigma instead for them.")
            variances[where_no_ref] = np.diag(self.covmat(ignore_external=True
                                                          ))[where_no_ref]
        return variances



"""
Testing some quantities not used yet by any internal likelihood.
"""

import pytest
import numpy as np
from copy import deepcopy

from cobaya.cosmo_input import planck_lss_precision, planck_base_model, create_input
from cobaya.model import get_model
from cobaya.tools import recursive_update, check_2d

from .test_cosmo_planck_2015 import params_lowTEB_highTTTEEE
from .conftest import install_test_wrapper
from .common import process_packages_path

fiducial_parameters = deepcopy(params_lowTEB_highTTTEEE)
redshifts = [100, 10, 1, 0]


def _get_model_with_requirements_and_eval(theo, reqs, packages_path, skip_not_installed):
    planck_base_model_prime = deepcopy(planck_base_model)
    planck_base_model_prime["hubble"] = "H"  # intercompatibility CAMB/CLASS
    info_theory = {theo: {"extra_args": planck_lss_precision[theo]}}
    info = create_input(planck_names=True, theory=theo, **planck_base_model_prime)
    info = recursive_update(info, {"theory": info_theory, "likelihood": {"one": None}})
    info["packages_path"] = process_packages_path(packages_path)
    info["debug"] = True
    model = install_test_wrapper(skip_not_installed, get_model, info)
    eval_parameters = {p: v for p, v in fiducial_parameters.items()
                       if p in model.parameterization.sampled_params()}
    model.add_requirements(reqs)
    model.logposterior(eval_parameters)
    return model


# sigma8(z), fsgima8(z) ##################################################################

sigma8_values = [0.0107213, 0.09646127, 0.5045227, 0.83073719]
fsigma8_values = [0.01063159, 0.09639173, 0.44311857, 0.43909749]


def _test_cosmo_sigma8_fsigma8(theo, packages_path, skip_not_installed):
    reqs = {"sigma8_z": {"z": redshifts}, "fsigma8": {"z": redshifts}}
    model = _get_model_with_requirements_and_eval(
        theo, reqs, packages_path, skip_not_installed)
    assert np.allclose(model.theory[theo].get_sigma8_z(redshifts),
                       sigma8_values, rtol=1e-5 if theo.lower() == "camb" else 5e-4)
    # NB: classy tolerance quite high for fsigma8!
    # (see also test of bao.sdss_dr16_baoplus_qso)
    assert np.allclose(model.theory[theo].get_fsigma8(redshifts),
                       fsigma8_values, rtol=1e-5 if theo.lower() == "camb" else 1e-2)


def test_cosmo_sigma8_fsigma8_camb(packages_path, skip_not_installed):
    _test_cosmo_sigma8_fsigma8("camb", packages_path, skip_not_installed)


def test_cosmo_sigma8_fsigma8_classy(packages_path, skip_not_installed):
    _test_cosmo_sigma8_fsigma8("classy", packages_path, skip_not_installed)


# sigma(R, z) ############################################################################

z_sigma_R = [0, 2, 5]
R_sigma_R = np.arange(1, 20, 1)
sigma_R_values = {
    ("delta_tot", "delta_tot"): [
        [2.91641221, 2.21110361, 1.83787574, 1.59298059, 1.41544224,
         1.27835804, 1.16812519, 1.07707886, 1.00019134, 0.93411662,
         0.87659859, 0.82600355, 0.78103409, 0.74077065, 0.70445364,
         0.67146121, 0.64136035, 0.61377842, 0.58839752],
        [1.21913872, 0.92428173, 0.76825062, 0.66586932, 0.59164694,
         0.53433678, 0.48825215, 0.45018873, 0.41804465, 0.39042106,
         0.36637479, 0.34522284, 0.3264228, 0.30959024, 0.2944076,
         0.28061489, 0.26803108, 0.2565004, 0.24588992],
        [0.61857867, 0.46896734, 0.38979628, 0.33784721, 0.30018608,
         0.27110628, 0.24772239, 0.22840853, 0.21209819, 0.19808161,
         0.18588019, 0.17514739, 0.16560799, 0.15706691, 0.14936304,
         0.14236443, 0.13597925, 0.13012844, 0.12474456]],
    ("delta_nonu", "delta_nonu"): [
        [2.92953248, 2.22099356, 1.84605073, 1.60002765, 1.42167022,
         1.2839528, 1.17321033, 1.08174289, 1.00449967, 0.93811928,
         0.88033523, 0.82950631, 0.7843291, 0.74387981, 0.70739533,
         0.67425091, 0.64401153, 0.61630285, 0.59080548],
        [1.22465058, 0.92844891, 0.77170492, 0.66885515, 0.59429274,
         0.53671975, 0.49042359, 0.45218534, 0.41989354, 0.39214295,
         0.36798608, 0.34673683, 0.3278503, 0.3109403, 0.29568781,
         0.28183165, 0.26918997, 0.25760626, 0.24694701],
        [0.62138387, 0.47109062, 0.39155828, 0.33937194, 0.30153863,
         0.27232581, 0.24883487, 0.22943255, 0.21304747, 0.19896663,
         0.18670925, 0.17592722, 0.16634405, 0.15776378, 0.15002454,
         0.14299381, 0.13657932, 0.13070165, 0.12529306]]}


def _test_cosmo_sigma_R(theo, packages_path, skip_not_installed):
    vars_pairs = (("delta_tot", "delta_tot"), ("delta_nonu", "delta_nonu"))
    reqs = {"sigma_R": {"z": z_sigma_R, "R": R_sigma_R, "vars_pairs": vars_pairs}}
    model = _get_model_with_requirements_and_eval(
        theo, reqs, packages_path, skip_not_installed)
    for pair in vars_pairs:
        z_out, R_out, sigma_R_out = model.theory[theo].get_sigma_R(pair)
        assert np.allclose(R_out, R_sigma_R)
        assert np.allclose(z_out, z_sigma_R)
        assert np.allclose(sigma_R_out, sigma_R_values[pair],
                           rtol=1e-5 if theo.lower() == "camb" else 2e-3)


def test_cosmo_sigma_R_camb(packages_path, skip_not_installed):
    _test_cosmo_sigma_R("camb", packages_path, skip_not_installed)


def test_cosmo_sigma_R_classy(packages_path, skip_not_installed):
    _test_cosmo_sigma_R("classy", packages_path, skip_not_installed)


# Omega_X(z) #############################################################################

Omega_b_values = [0.15172485, 0.15517809, 0.12258897, 0.04920226]
Omega_cdm_values = [0.81730093, 0.83590262, 0.66035382, 0.26503934]
Omega_nu_massive_values = [0.00608623, 0.0045243, 0.00355319, 0.00142589]


def _test_cosmo_omega(theo, packages_path, skip_not_installed):
    reqs = {"Omega_b": {"z": redshifts}, "Omega_cdm": {"z": redshifts},
            "Omega_nu_massive": {"z": redshifts}}
    model = _get_model_with_requirements_and_eval(
        theo, reqs, packages_path, skip_not_installed)
    assert np.allclose(model.theory[theo].get_Omega_b(redshifts),
                       Omega_b_values, rtol=1e-5 if theo.lower() == "camb" else 5e-4)
    assert np.allclose(model.theory[theo].get_Omega_cdm(redshifts),
                       Omega_cdm_values, rtol=1e-5 if theo.lower() == "camb" else 5e-4)
    assert np.allclose(model.theory[theo].get_Omega_nu_massive(redshifts),
                       Omega_nu_massive_values,
                       rtol=1e-5 if theo.lower() == "camb" else 2e-3)


def test_cosmo_omega_camb(packages_path, skip_not_installed):
    _test_cosmo_omega("camb", packages_path, skip_not_installed)


def test_cosmo_omega_classy(packages_path, skip_not_installed):
    _test_cosmo_omega("classy", packages_path, skip_not_installed)


# angular_diameter_distance_2 ############################################################

ang_diam_dist_2_values = [
    31.59567987, 93.34513188, 127.08027199, 566.97224099, 876.72216398, 1703.62457558]


def _test_cosmo_ang_diam_dist_2(theo, packages_path, skip_not_installed):
    reqs = {"angular_diameter_distance_2": {"z_pairs": redshifts}}
    model = _get_model_with_requirements_and_eval(
        theo, reqs, packages_path, skip_not_installed)
    redshift_pairs = check_2d(redshifts)
    assert np.allclose(model.theory[theo].get_angular_diameter_distance_2(redshift_pairs),
                       ang_diam_dist_2_values, rtol=1e-5)


def test_cosmo_ang_diam_dist_2_camb(packages_path, skip_not_installed):
    _test_cosmo_ang_diam_dist_2("camb", packages_path, skip_not_installed)


def test_cosmo_ang_diam_dist_2_classy(packages_path, skip_not_installed):
    _test_cosmo_ang_diam_dist_2("classy", packages_path, skip_not_installed)


# Weyl power spectrum ####################################################################

var_pair = ("Weyl", "Weyl")
zs = [0, 0.5, 1, 1.5, 2, 3.5]
ks = np.logspace(-4, np.log10(15), 5)
Pkz_values = [[-27.43930515, -24.66945975, -24.57476343, -28.00109372, -33.24474829],
              [-27.15442496, -24.3843334, -24.28538885, -28.05600287, -33.27033893],
              [-27.05300137, -24.28273214, -24.1803611, -28.25169186, -33.31505885],
              [-27.01121008, -24.24079765, -24.13585631, -28.46595848, -33.3742216],
              [-26.99148443, -24.22094486, -24.11405468, -28.65842103, -33.44075841],
              [-26.97141436, -24.20052295, -24.08988483, -29.05073038, -33.70090064]]


def _test_cosmo_weyl_pkz(theo, packages_path, skip_not_installed):
    # Similar to what"s requested by DES (but not used by default)
    reqs = {"Pk_interpolator": {"z": zs, "k_max": max(ks), "nonlinear": (False, True),
                                "vars_pairs": var_pair}}
    model = _get_model_with_requirements_and_eval(
        theo, reqs, packages_path, skip_not_installed)
    interp = model.theory[theo].get_Pk_interpolator(var_pair=var_pair, nonlinear=True)
    assert np.allclose(Pkz_values, interp.logP(zs, ks),
                       rtol=1e-5 if theo.lower() == "camb" else 5e-4)


def test_cosmo_weyl_pkz_camb(packages_path, skip_not_installed):
    _test_cosmo_weyl_pkz("camb", packages_path, skip_not_installed)


@pytest.mark.skip
def test_cosmo_weyl_pkz_classy(packages_path, skip_not_installed):
    _test_cosmo_weyl_pkz("classy", packages_path, skip_not_installed)



r"""
.. module:: bao

:Synopsis: BAO, f_sigma8 and other measurements at single redshifts, with correlations
:Author: Antony Lewis, Pablo Lemos (adapted to Cobaya by Jesus Torrado, with little
        modification)

This code provides a template for BAO, :math:`f\sigma_8`, :math:`H`
and other redshift dependent functions.

The datasets implemented at this moment are:

- ``bao.sixdf_2011_bao``
- ``bao.sdss_dr7_mgs``
- ``bao.sdss_dr12_consensus_bao``
- ``bao.sdss_dr12_consensus_full_shape``
- ``bao.sdss_dr12_consensus_final`` (combined data of the previous two)
- ``bao.sdss_dr16_baoplus_lrg`` (combining data from BOSS DR12 and eBOSS DR16)
- ``bao.sdss_dr16_baoplus_lyauto``
- ``bao.sdss_dr16_baoplus_lyxqso``
- ``bao.sdss_dr16_baoplus_qso``
- ``bao.sdss_dr12_lrg_bao_dmdh``
- ``bao.sdss_dr16_lrg_bao_dmdh``


.. |br| raw:: html

   <br />

.. note::

   If you use the likelihood ``sixdf_2011_bao``, please cite:
   |br|
   F. Beutler et al,
   `The 6dF Galaxy Survey: baryon acoustic oscillations and the local Hubble constant`
   `(arXiv:1106.3366) <https://arxiv.org/abs/1106.3366>`_
   |br| |br|
   If you use the likelihood ``sdss_dr7_mgs``, please cite:
   |br|
   A.J. Ross et al,
   `The clustering of the SDSS DR7 main Galaxy sample - I.
   A 4 per cent distance measure at z = 0.15`
   `(arXiv:1409.3242) <https://arxiv.org/abs/1409.3242>`_
   |br| |br|
   If you use any of the likelihoods ``sdss_dr12_*``, please cite:
   |br|
   S. Alam et al,
   `The clustering of galaxies in the completed SDSS-III Baryon Oscillation Spectroscopic
   Survey: cosmological analysis of the DR12 galaxy sample`
   `(arXiv:1607.03155) <https://arxiv.org/abs/1607.03155>`_
   |br| |br|
   If you use any of the likelihoods ``sdss_dr16_*``, please cite:
   |br|
   S. Alam et al,
   `The Completed SDSS-IV extended Baryon Oscillation Spectroscopic Survey: Cosmological
   Implications from two Decades of Spectroscopic Surveys at the Apache Point observatory`
   `(arXiv:2007.08991) <https://arxiv.org/abs/2007.08991>`_


Usage
-----

To use any of these likelihoods, simply mention them in the likelihoods block, or add them
using the :doc:`input generator <cosmo_basic_runs>`.

These likelihoods have no nuisance parameters or particular settings that you may want
to change (except for the installation path; see below).

Note that although called "bao", many of these data combinations also include redshift
distortion data (RSD), encapsulated via a single "f sigma8" parameter (which is not
accurate for some non-LCDM models).


Defining your own BAO likelihood
--------------------------------

You can use the likelihood ``bao.generic`` as a template for any BAO data.

To do that, create a file containing the data points, e.g. ``myBAO.dat``, as

.. code::

   [z] [value at z] [quantity]
   ...

where you can use as many different quantities and redshifts as you like.

The available quantities are

- ``DV_over_rs``: Spherically-averaged distance, over sound horizon radius
- ``rs_over_DV``: Idem, inverse
- ``DM_over_rs``: Comoving angular diameter distance, over sound horizon radius
- ``DA_over_rs``: Physical angular diameter distance, over sound horizon radius
- ``Hz_rs``: Hubble parameter, times sound horizon radius
- ``f_sigma8``: Differential matter linear growth rate,
  times amplitude of present-day fuctuations
- ``F_AP``: Anisotropy (Alcock-Paczynski) parameter

In addition create a file, e.g. ``myBAO.cov``, containing the covariance matrix for those
data, with the same row order as the data file.

Now, add to your likelihood block:

.. literalinclude:: ../cobaya/likelihoods/bao/generic.yaml
   :language: yaml

You can rename your BAO likelihood and use multiple ones with different data (see
:ref:`likelihood_rename`).


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

Manual installation of the BAO likelihood data
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Assuming you are installing all your
likelihoods under ``/path/to/likelihoods``, simply do

.. code:: bash

   $ cd /path/to/likelihoods
   $ git clone https://github.com/CobayaSampler/bao_data.git

After this, mention the path to this likelihood when you include it in an input file, e.g.

.. code-block:: yaml

   likelihood:
     bao.sdss_dr12_consensus_[bao|full_shape|final|...]:
       path: /path/to/likelihoods/bao_data
"""
# Global
import os
import numpy as np
from scipy.interpolate import UnivariateSpline, RectBivariateSpline, \
    RegularGridInterpolator
import pandas as pd
from typing import Optional, Sequence

# Local
from cobaya.log import LoggedError
from cobaya.conventions import Const, packages_path_input
from cobaya.likelihoods.base_classes import InstallableLikelihood


class BAO(InstallableLikelihood):
    # Data type for aggregated chi2 (case sensitive)
    type = "BAO"

    install_options = {"github_repository": "CobayaSampler/bao_data",
                       "github_release": "v2.1"}

    prob_dist_bounds: Optional[Sequence[float]]
    measurements_file: Optional[str] = None
    rs_fid: Optional[float] = None
    rs_rescale: Optional[float] = None
    prob_dist: Optional[str] = None
    cov_file: Optional[str] = None
    invcov_file: Optional[str] = None
    redshift: Optional[float] = None
    observable_1: Optional[str] = None
    observable_2: Optional[str] = None
    observable_3: Optional[str] = None
    grid_file: Optional[str] = None
    path: Optional[str]

    def initialize(self):
        if not getattr(self, "path", None) and \
                not getattr(self, packages_path_input, None):
            raise LoggedError(
                self.log, "No path given to BAO data. Set the likelihood property "
                          "'path' or the common property '%s'.", packages_path_input)
        # If no path specified, use the external packages path
        data_file_path = os.path.normpath(getattr(self, "path", None) or
                                          os.path.join(self.packages_path, "data"))
        # Rescaling by a fiducial value of the sound horizon
        if self.rs_rescale is None:
            if self.rs_fid is not None:
                self.rs_rescale = 1 / self.rs_fid
            else:
                self.rs_rescale = 1
        # Load "measurements file" and covmat of requested
        if self.measurements_file:
            try:
                self.data = pd.read_csv(
                    os.path.join(data_file_path, self.measurements_file),
                    header=None, index_col=None, sep=r"\s+", comment="#")
            except IOError:
                raise LoggedError(
                    self.log, "Couldn't find measurements file '%s' in folder '%s'. " % (
                        self.measurements_file, data_file_path) + "Check your paths.")
        elif self.grid_file:
            pass
        else:
            self.data = pd.DataFrame([self.data] if not hasattr(self.data[0], "__len__")
                                     else self.data)

        if not self.grid_file:
            self.use_grid_2d = False
            self.use_grid_3d = False
            # Columns: z value [err] [type]
            self.has_type = self.data.iloc[:, -1].dtype == np.dtype("O")
            assert self.has_type  # mandatory for now!
            self.has_err = len(self.data.columns) > 2 and self.data[2].dtype == float
            if self.has_err:
                self.data.columns = ["z", "value", "error", "observable"]
            else:
                self.data.columns = ["z", "value", "observable"]
            prefix = "bao_"
            self.data["observable"] = [(c[len(prefix):] if c.startswith(prefix) else c)
                                       for c in self.data["observable"]]

        # Probability distribution
        if self.prob_dist:
            try:
                chi2 = np.loadtxt(os.path.join(data_file_path, self.prob_dist))
            except IOError:
                raise LoggedError(
                    self.log, "Couldn't find probability distribution file '%s' "
                              "in folder '%s'. " % (self.prob_dist, data_file_path) +
                              "Check your paths.")
            try:
                assert self.prob_dist_bounds
                alpha = np.linspace(
                    self.prob_dist_bounds[0], self.prob_dist_bounds[1], len(chi2))
            except (TypeError, AttributeError, IndexError, ValueError):
                raise LoggedError(
                    self.log, "If 'prob_dist' given, 'prob_dist_bounds' needs to be "
                              "specified as [min, max].")
            spline = UnivariateSpline(alpha, -chi2 / 2, s=0)
            self.logpdf = lambda _x: (spline(_x)[0] if self.prob_dist_bounds[0]
                                                       <= _x <= self.prob_dist_bounds[1]
                                      else -np.inf)
        elif self.grid_file:
            try:
                self.grid_data = np.loadtxt(
                    os.path.join(data_file_path, self.grid_file))
            except IOError:
                raise LoggedError(
                    self.log, "Couldn't find grid file '%s' in folder '%s'. " % (
                        self.grid_file, data_file_path) + "Check your paths.")
            if not self.observable_1 or not self.observable_2:
                raise LoggedError(
                    self.log, "If using grid data, 'observable_1' and 'observable_2'"
                              "need to be specified.")
            if self.redshift is None:
                raise LoggedError(
                    self.log, "If using grid data, 'redshift'"
                              "needs to be specified.")

            self.has_type = True  # Not sure what this is
            self.data = pd.DataFrame()

            if self.grid_data.shape[1] == 3:
                self.use_grid_2d = True
                self.use_grid_3d = False
                self.data["observable"] = [self.observable_1, self.observable_2]

                x = np.unique(self.grid_data[:, 0])
                y = np.unique(self.grid_data[:, 1])

                Nx = x.shape[0]
                Ny = y.shape[0]

                chi2 = np.reshape(np.log(self.grid_data[:, 2]), [Nx, Ny])

                # Make the interpolator (x refers to at, y refers to ap).
                self.interpolator = RectBivariateSpline(x, y, chi2, kx=3, ky=3)
            elif self.grid_data.shape[1] == 4:
                self.use_grid_2d = False
                self.use_grid_3d = True
                if not self.observable_3:
                    raise LoggedError(
                        self.log, "If using 3D grid data, 'observable_3'"
                                  "needs to be specified.")
                self.data["observable"] = [self.observable_1, self.observable_2,
                                           self.observable_3]

                x = np.unique(self.grid_data[:, 0])
                y = np.unique(self.grid_data[:, 1])
                z = np.unique(self.grid_data[:, 2])

                Nx = x.shape[0]
                Ny = y.shape[0]
                Nz = z.shape[0]

                chi2 = np.reshape(np.log(self.grid_data[:, 3] + 1e-300), [Nx, Ny, Nz])

                self.interpolator3D = RegularGridInterpolator((x, y, z), chi2,
                                                              bounds_error=False,
                                                              fill_value=np.log(1e-300))

            else:
                raise LoggedError(
                    self.log, "Grid data has the wrong dimensions")
                # Covariance --> read and re-sort as self.data
        else:
            self.use_grid_2d = False
            self.use_grid_3d = False
            try:
                if self.cov_file:
                    self.cov = np.loadtxt(os.path.join(data_file_path, self.cov_file))
                elif self.invcov_file:
                    invcov = np.loadtxt(os.path.join(data_file_path, self.invcov_file))
                    self.cov = np.linalg.inv(invcov)
                elif "error" in self.data.columns:
                    self.cov = np.diag(self.data["error"] ** 2)
                else:
                    raise LoggedError(
                        self.log, "No errors provided, either as cov, invcov "
                                  "or as the 3rd column in the data file.")
                self.invcov = np.linalg.inv(self.cov)
            except IOError:
                raise LoggedError(
                    self.log, "Couldn't find (inv)cov file '%s' in folder '%s'. " % (
                        self.cov_file or self.invcov_file,
                        data_file_path) + "Check your paths.")
            self.logpdf = lambda _x: (lambda x_: -0.5 * x_.dot(self.invcov).dot(x_))(
                _x - self.data["value"].values)
            self.log.info("Initialized.")

    def get_requirements(self):
        # Requisites
        if self.use_grid_2d:
            zs = {self.observable_1: np.array([self.redshift]),
                  self.observable_2: np.array([self.redshift])}
        elif self.use_grid_3d:
            zs = {self.observable_1: np.array([self.redshift]),
                  self.observable_2: np.array([self.redshift]),
                  self.observable_3: np.array([self.redshift])
                  }
        else:
            zs = {obs: self.data.loc[self.data["observable"] == obs, "z"].values
                  for obs in self.data["observable"].unique()}
        theory_reqs = {
            "DV_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DV_over_rs", None)},
                "Hubble": {"z": zs.get("DV_over_rs", None)},
                "rdrag": None},
            "rs_over_DV": {
                "angular_diameter_distance": {"z": zs.get("rs_over_DV", None)},
                "Hubble": {"z": zs.get("rs_over_DV", None)},
                "rdrag": None},
            "DM_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DM_over_rs", None)},
                "rdrag": None},
            "DA_over_rs": {
                "angular_diameter_distance": {"z": zs.get("DA_over_rs", None)},
                "rdrag": None},
            "DH_over_rs": {
                "Hubble": {"z": zs.get("DH_over_rs", None)},
                "rdrag": None},
            "Hz_rs": {
                "Hubble": {"z": zs.get("Hz_rs", None)},
                "rdrag": None},
            "f_sigma8": {
                "fsigma8": {"z": zs.get("f_sigma8", None)},
            },
            "F_AP": {
                "angular_diameter_distance": {"z": zs.get("F_AP", None)},
                "Hubble": {"z": zs.get("F_AP", None)}}}
        obs_used_not_implemented = np.unique([obs for obs in self.data["observable"]
                                              if obs not in theory_reqs])
        if len(obs_used_not_implemented):
            raise LoggedError(
                self.log, "This likelihood refers to observables '%s' that have not been"
                          " implemented yet. Did you mean any of %s? "
                          "If you didn't, please, open an issue in github.",
                obs_used_not_implemented, list(theory_reqs))
        requisites = {}
        if self.has_type:
            for obs in self.data["observable"].unique():
                requisites.update(theory_reqs[obs])
        return requisites

    def theory_fun(self, z, observable):
        # Functions to get the corresponding theoretical prediction:
        # Spherically-averaged distance, over sound horizon radius
        if observable == "DV_over_rs":
            return np.cbrt(
                ((1 + z) * self.provider.get_angular_diameter_distance(z)) ** 2 *
                Const.c_km_s * z / self.provider.get_Hubble(z,
                                                            units="km/s/Mpc")) / self.rs()
        # Idem, inverse
        elif observable == "rs_over_DV":
            return np.cbrt(
                ((1 + z) * self.provider.get_angular_diameter_distance(z)) ** 2 *
                Const.c_km_s * z / self.provider.get_Hubble(z, units="km/s/Mpc")) ** (
                       -1) * self.rs()
        # Comoving angular diameter distance, over sound horizon radius
        elif observable == "DM_over_rs":
            return (1 + z) * self.provider.get_angular_diameter_distance(z) / self.rs()
        # Physical angular diameter distance, over sound horizon radius
        elif observable == "DA_over_rs":
            return self.provider.get_angular_diameter_distance(z) / self.rs()
        # Hubble distance [c/H(z)] over sound horizon radius.
        elif observable == "DH_over_rs":
            return 1 / self.provider.get_Hubble(z, units="1/Mpc") / self.rs()
        # Hubble parameter, times sound horizon radius
        elif observable == "Hz_rs":
            return self.provider.get_Hubble(z, units="km/s/Mpc") * self.rs()
        # Diff Linear Growth Rate times present amplitude
        elif observable == "f_sigma8":
            return self.provider.get_fsigma8(z)
        # Anisotropy (Alcock-Paczynski) parameter
        elif observable == "F_AP":
            return ((1 + z) * self.provider.get_angular_diameter_distance(z) *
                    self.provider.get_Hubble(z, units="km/s/Mpc")) / Const.c_km_s

    def rs(self):
        return self.provider.get_param("rdrag") * self.rs_rescale

    def logp(self, **params_values):
        if self.use_grid_2d:
            x = self.theory_fun(self.redshift, self.observable_1)
            y = self.theory_fun(self.redshift, self.observable_2)
            chi2 = float(self.interpolator(x, y)[0])
            return chi2
        elif self.use_grid_3d:
            x = self.theory_fun(self.redshift, self.observable_1)
            y = self.theory_fun(self.redshift, self.observable_2)
            z = self.theory_fun(self.redshift, self.observable_3)
            chi2 = self.interpolator3D(np.array([x, y, z])[:, 0])
            return chi2
        else:
            theory = np.array([self.theory_fun(z, obs) for z, obs
                               in zip(self.data["z"], self.data["observable"])]).T[0]
            if self.is_debug():
                for i, (z, obs, theo) in enumerate(
                        zip(self.data["z"], self.data["observable"], theory)):
                    self.log.debug("%s at z=%g : %g (theo) ; %g (data)",
                                   obs, z, theo, self.data.iloc[i, 1])
            return self.logpdf(theory)



from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE_lite(Planck2018Clik):
    r"""
    High-$\ell$ temperature and polarization, foreground-marginalized \textsc{plik\_lite}
    likelihood of Planck's 2018 data release \cite{Aghanim:2019ame}.
    """
    pass



from cobaya.likelihoods.base_classes import PlanckPlikLite


class TT_lite_native(PlanckPlikLite):
    r"""
    High-$\ell$ temperature-only, foreground-marginalized likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame} (Python re-implementation of the official
    \textsc{plik\_lite}, by E.~Calabrese, and A.~Lewis).
    """
    pass
